
from .celery_main import app as celery_app
from django.http import JsonResponse

import logging

logger = logging.getLogger('default')


def background_task(*args, **kwargs):
    return run_task.delay(*args, **kwargs)


def background_status(request, task_id):
    task = celery_app.AsyncResult(id=task_id)
    rslt = task.result
    rslt.update({'task_id': task_id})
    return JsonResponse(rslt)


@celery_app.task(bind=True)
def run_task(self, *args, **kwargs):
    task_name = kwargs['task']
    self.update_state(state='PROGRESS', meta={'status': 'IN_PROGRESS', 'task_name': task_name})

    try:
        rslts = kwargs['calc'](*args, **kwargs)
        return {'status': 'COMPLETED', 'task_name': task_name, 'output': rslts}
    except Exception as e:
        return {'status': 'FAILED', 'task_name': task_name, 'output': {'error': str(e)}}
    


import urllib.request, urllib.error, urllib.parse
import json

from accountifie.common.api import api_func



def lpq():
    base_url = api_func('environment', 'variable', 'ACCOUNTIFIE_SVC_URL')
    url = '%s/lpq/stats' % base_url
    request = urllib.request.Request(url)
    response = urllib.request.urlopen(request)
    json_result = json.load(response)
    return json_result


def gl_stats(company_id):
    base_url = api_func('environment', 'variable', 'ACCOUNTIFIE_SVC_URL')
    url = '%s/gl/%s/stats' % (base_url, company_id)
    request = urllib.request.Request(url)
    response = urllib.request.urlopen(request)
    json_result = json.load(response)
    return json_result



import os
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.db.models import get_models
from django.db import transaction

from accountifie.gl.bmo import BusinessModelObject
from accountifie.gl.models import Transaction, TranLine
from query.query_manager_strategy_factory import QueryManagerStrategyFactory

class Command(BaseCommand):

    help = 'Rebuild ALL general ledger entries'
    
    def handle(self, *args, **options):
        verbosity = options["verbosity"]
        if verbosity > 0:
            print("Dropping general ledger and rebuilding all entries")

        with transaction.atomic():
            #quicker to do the two delets ourselves than let the engine work out all the cascading stuff
            TranLine.objects.all().delete()
            #Transaction.objects.all().delete()
            QueryManagerStrategyFactory().erase('*')


        QueryManagerStrategyFactory().set_fast_inserts('*', True)
        for model in get_models():
            
            if issubclass(model, BusinessModelObject):
                if verbosity > 0:
                    print("rebuilding GL entries from %s.%s" % (model._meta.app_label, model.__name__))
                with transaction.atomic():
                    for bmo in model.objects.all():
                        bmo.save()
        QueryManagerStrategyFactory().set_fast_inserts('*', False)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0006_auto_20171221_0736'),
    ]

    operations = [
        migrations.AddField(
            model_name='company',
            name='label',
            field=models.CharField(default=b'TBD', max_length=20),
        ),
        migrations.AddField(
            model_name='counterparty',
            name='label',
            field=models.CharField(default=b'TBD', max_length=20),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-28 03:40
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('gl', '0002_transaction_bmo_id'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='transaction',
            name='company',
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='tranline',
            name='transaction',
        ),
        migrations.AddField(
            model_name='tranline',
            name='bmo_id',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='tranline',
            name='comment',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='tranline',
            name='company',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='gl.Company'),
        ),
        migrations.AddField(
            model_name='tranline',
            name='content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='tranline',
            name='date_end',
            field=models.DateField(blank=True, db_index=True, null=True),
        ),
        migrations.AddField(
            model_name='tranline',
            name='object_id',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='account',
            name='display_name',
            field=models.CharField(blank=True, help_text='the name you would use on a report title', max_length=100),
        ),
        migrations.AlterField(
            model_name='account',
            name='id',
            field=models.CharField(help_text='Unique short code or number', max_length=20, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='account',
            name='path',
            field=models.CharField(db_index=True, help_text='dotted-path reference to show location on balance sheet', max_length=100, unique=True),
        ),
        migrations.AlterField(
            model_name='account',
            name='role',
            field=models.CharField(choices=[('asset', 'asset'), ('liability', 'liability'), ('income', 'income'), ('expense', 'expense'), ('capital', 'capital')], help_text='used in P&L queries and for picklists', max_length=10),
        ),
        migrations.AlterField(
            model_name='company',
            name='cmpy_type',
            field=models.CharField(choices=[('ALO', 'Standalone'), ('CON', 'Consolidation')], max_length=10),
        ),
        migrations.AlterField(
            model_name='tranline',
            name='account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='gl.Account'),
        ),
        migrations.DeleteModel(
            name='Transaction',
        ),
    ]



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='transaction',
            name='bmo_id',
            field=models.CharField(default='', max_length=100),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-27 01:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0003_address_label'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskmonitor',
            name='task_state',
            field=models.CharField(choices=[('started', 'started'), ('failed', 'failed'), ('succeeded', 'succeeded')], max_length=200, null=True),
        ),
    ]



"""
Strategy that delegates transaction querying to accountifie-svc.

The new kid on the block, capable of going from 0 to 100km/h in under 50ms
though with the chance of a few early speed wobbles.

See query_manager_strategy.py for interface docs.
Use query_manager_strategy_factory.py to get an instance of this class.
"""

import pandas as pd
import itertools
import json

try: #python3
    from urllib.request import urlopen
except: #python2
    from urllib2 import urlopen

from decimal import Decimal
from collections import defaultdict

from dateutil.parser import parse
from django.conf import settings
from .query_manager_strategy import QueryManagerStrategy
from accountifie.common.api import api_func
import logging

import time
DZERO = Decimal('0')
logger = logging.getLogger('default')


class QueryManagerRemoteStrategy(QueryManagerStrategy):
    def cp_balances_for_dates(self, company_id, account_ids, dates):        
        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.cp_balances_for_dates(cmpny, account_ids, dates) for cmpny in company_list]
            return self.__merge_cp_balances_for_dates_results(balances)
        
        client = accountifieSvcClient(company_id)
        result = defaultdict(dict)

        for dt in dates:
            start = dates[dt]['start']
            end = dates[dt]['end']
            cp_balances = client.cp_balances(accounts=account_ids,
                                             from_date=start,
                                             to_date=end)

            for acct_bals in cp_balances:
                result[acct_bals['id']][dt] = {'openingBalance': acct_bals['openingBalance'],
                                               'closingBalance': acct_bals['closingBalance']}

        return result

    def account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra, with_tags, excl_tags):
        
        interco_exempt_accounts = api_func('gl', 'externalaccounts')

        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.account_balances_for_dates(cmpny, account_ids, dates, with_counterparties, True, excl_contra, with_tags, excl_tags) for cmpny in company_list]
            return self.__merge_account_balances_for_dates_results(balances)

        
        client = accountifieSvcClient(company_id)
        date_indexed_account_balances = {}

        interco_counterparties = self.get_interco_counterparties_for(company_id) if excl_interco else []
        accounts_with_interco_exclusion =    [id for id in account_ids if id not in interco_exempt_accounts] if excl_interco else []
        accounts_without_interco_exclusion = [id for id in account_ids if id in interco_exempt_accounts]     if excl_interco else account_ids

        for dt in dates:
            start = dates[dt]['start']
            end = dates[dt]['end']

            interco_excluded_balances = client.balances(
                accounts=accounts_with_interco_exclusion,
                from_date=start,
                to_date=end,
                with_counterparties=with_counterparties,
                excluding_counterparties=interco_counterparties,
                excluding_contra_accounts=excl_contra,
                with_tags=with_tags,
                excluding_tags=excl_tags
            ) if len(accounts_with_interco_exclusion) > 0 else []

            interco_included_balances = client.balances(
                accounts=accounts_without_interco_exclusion,
                from_date=start,
                to_date=end,
                with_counterparties=with_counterparties,
                excluding_contra_accounts=excl_contra,
                with_tags=with_tags,
                excluding_tags=excl_tags
            ) if len(accounts_without_interco_exclusion) > 0 else []

            balances = interco_excluded_balances + interco_included_balances
            date_indexed_account_balances[dt] = dict((balance['id'], float(balance['shift'])) for balance in balances)        
        return date_indexed_account_balances

    def transactions(self, company_id, account_ids, from_date, to_date, chunk_frequency, with_counterparties, excl_interco, excl_contra):
        interco_exempt_accounts = api_func('gl', 'externalaccounts')

        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.transactions(cmpny, account_ids, from_date, to_date, chunk_frequency, with_counterparties, True, excl_contra) for cmpny in company_list]
            return self.__merge_transactions_results(balances)

        client = accountifieSvcClient(company_id)

        interco_counterparties = self.get_interco_counterparties_for(company_id) if excl_interco else []
        accounts_with_interco_exclusion =    [id for id in account_ids if id not in interco_exempt_accounts] if excl_interco else []
        accounts_without_interco_exclusion = [id for id in account_ids if id in interco_exempt_accounts]     if excl_interco else account_ids

        interco_excluded_transactions = client.transactions(
            accounts=accounts_with_interco_exclusion,
            from_date=from_date,
            to_date=to_date,
            chunk_frequency=chunk_frequency,
            excluding_counterparties=interco_counterparties,
            with_counterparties=with_counterparties,
            excluding_contra_accounts=excl_contra
        ) if len(accounts_with_interco_exclusion) > 0 else {}

        interco_included_transactions = client.transactions(
            accounts=accounts_without_interco_exclusion,
            from_date=from_date,
            to_date=to_date,
            chunk_frequency=chunk_frequency,
            with_counterparties=with_counterparties,
            excluding_contra_accounts=excl_contra
        ) if len(accounts_without_interco_exclusion) > 0 else {}

        # merge the two dictionaries
        transactions = interco_excluded_transactions.copy()
        transactions.update(interco_included_transactions)

        make_contras = lambda x: 'more than two' if len(x)>2 else '.'.join(x)

        formatted_transactions = []
        for account in transactions:
            for transaction in transactions[account]:
                if transaction['amount'] != '0.00':
                    formatted_transactions.append({
                        'date': parse(transaction['dateEnd'] if ('dateEnd' in transaction and transaction['dateEnd']) else transaction['date']).date(),
                        'id': transaction['id'],
                        'comment': transaction['comment'],
                        'account_id': account,
                        'contra_accts': make_contras(transaction['contraAccounts']),
                        'counterparty': transaction['counterparty'],
                        'amount': float(transaction['amount'])
                    })

        return formatted_transactions

    def __merge_cp_balances_for_dates_results(self, result_list):
        merged_results = {}

        for result in result_list:
            for acct in result:
                merged_results[acct] = merged_results[acct] if acct in merged_results else {}
                for date in result[acct]:
                    if date not in merged_results[acct]:
                        merged_results[acct][date] = {'closingBalance': [], 'openingBalance': []}
                    existing_amount = merged_results[acct][date]
                    for col in ['closingBalance', 'openingBalance']:
                        all_rows = existing_amount[col] + result[acct][date][col]
                        merged_results[acct][date][col] = []
                        for k,v in itertools.groupby(sorted(all_rows, key=lambda x: x['cp']), key=lambda x: x['cp']):
                            merged_results[acct][date][col].append({'cp': k, 'total': str(sum(Decimal(r['total']) for r in v))})
        return merged_results

    def __merge_account_balances_for_dates_results(self, result_list):
        merged_results = {}

        for result in result_list:
            for date in result:
                merged_results[date] = merged_results[date] if date in merged_results else {}
                for account in result[date]:
                    existing_amount = merged_results[date][account] if account in merged_results[date] else float(0)
                    merged_results[date][account] = existing_amount + result[date][account]

        return merged_results

    def __merge_transactions_results(self, result_list):
        return [transaction for result in result_list for transaction in result]

    def get_transaction(self, company_id, transaction_id):
        client = accountifieSvcClient(company_id)
        return client.get_transaction(transaction_id)

    def create_gl_transactions(self, d2, lines, trans_id, bmo_id):
        if sum([Decimal(line['amount']) for line in lines]) == DZERO:
            self.upsert_transaction({
                'id': trans_id,
                'bmo_id': bmo_id,
                'date': str(d2['date']),
                'date_end': str(d2.get('date_end', None) or d2['date']),
                'comment': d2['comment'],
                'company': d2['company'],
                'type': None,
                'lines': [dict(account=line['account'],
                                amount=str(Decimal(line['amount'])),
                                counterparty = line['counterparty'],
                                tags = line['tags'] ) for line in lines]
            })
        else:
            logger.info('Unbalanced entry for bmo ID: %s' % bmo_id)


    def upsert_transaction(self, transaction):
        client = accountifieSvcClient(transaction['company'])
        client.upsert_transaction(transaction)

    def delete_transaction(self, company_id, transaction_id):
        client = accountifieSvcClient(company_id)
        client.delete_transaction(transaction_id)

    def delete_bmo_transactions(self, company_id, bmo_id):
        client = accountifieSvcClient(company_id)
        client.delete_bmo_transactions(bmo_id)

    def take_snapshot(self, company_id, snapshot_time=None):
        accountifieSvcClient(company_id).take_snapshot(snapshot_time)

    def erase(self, company_id):
        accountifieSvcClient(company_id).erase()

    def set_fast_inserts(self, company_id, value):
        if value:
            accountifieSvcClient(company_id).disable_balance_cache()
        else:
            accountifieSvcClient(company_id).enable_balance_cache()


    @staticmethod
    def __inter_co(row):
        ext_accts = api_func('gl', 'externalaccounts')
        companies = [cmpy['id'] for cmpy in api_func('gl', 'company') if cmpy['id']!=company]
        if row['account_id'] in ext_accts:
            return False
        if row['counterparty'] in companies:
            return True
        else:
            return False

    @staticmethod
    def get_interco_counterparties_for(company):
        output = [cmpy['id'] for cmpy in api_func('gl', 'company') if cmpy['id']!=company and cmpy['cmpy_type']=='ALO']
        output += [x.lower() for x in output]
        return output

class accountifieSvcClient(object):
    def __init__(self, company_id='*'):
        self.company_id = company_id
        self.url_base = "%s/gl/%s" % (self.__accountifie_svc_url(), company_id)

    def __post(self, path, params=None):
        params = json.dumps(params) if params is not None else '{}'
        url = '%s%s' % (self.url_base, path)
        request = urllib.request.Request(url, data=params, headers={'Content-Type': 'application/json'})
        try:
            response = urlopen(request)
            json_result = json.load(response)
            return json_result
        except:
            logger.error('Accountifie svc post failed on %s with params %s', url, str(params))
            return None

    def __get(self, path, params=None):
        if params:
            params = urllib.parse.urlencode(params)
            url = '%s%s?%s' % (self.url_base, path, params)
        else:
            url = '%s%s' % (self.url_base, path)

        start_time = time.time()
        request = urllib.request.Request(url)
        response = urlopen(request)
        json_result = json.load(response)

        resp_time = time.time() - start_time
        if resp_time > 2:
            logger.info('Accountifie-svc. Long response time. %.2f seconds. URL=%s' % (resp_time, url))

        return json_result

    def balances(self, accounts, from_date=None, to_date=None, with_counterparties=None, excluding_counterparties=None, excluding_contra_accounts=None, with_tags=None, excluding_tags=None):
        from_date = None if from_date == '2000-01-01' else from_date
        account_balances = self.__get('/balances', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date,
            'withCounterparties': ','.join(with_counterparties or []),
            'excludingCounterparties': ','.join(excluding_counterparties or []),
            'excludingContraAccounts': ','.join(excluding_contra_accounts or []),
            'withTags': ','.join(with_tags or []),
            'excludingTags': ','.join(excluding_tags or []),
        })

        return account_balances

    def cp_balances(self, accounts, from_date=None, to_date=None):
        from_date = None if from_date == '2000-01-01' else from_date
        cp_balances = self.__get('/cpBalances', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date
        })

        return cp_balances

    def transactions(self, accounts, from_date=None, to_date=None, chunk_frequency=None, with_counterparties=None, excluding_counterparties=None, excluding_contra_accounts=None):
        from_date = None if from_date == '2000-01-01' else from_date
        transaction_history = self.__get('/transactions', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date,
            'chunkFrequency': chunk_frequency,
            'withCounterparties': ','.join(with_counterparties or []),
            'excludingCounterparties': ','.join(excluding_counterparties or []),
            'excludingContraAccounts': ','.join(excluding_contra_accounts or [])
        })
        return transaction_history


    def get_transaction(self, transaction_id):
        return self.__get('/transaction/%s' % transaction_id)


    def upsert_transaction(self, transaction):
        self.__post('/transaction/%s/upsert' % transaction['id'], {
            'id': transaction['id'],
            'bmoId': transaction['bmo_id'],
            'date': transaction['date'],
            'dateEnd': transaction['date_end'],
            'comment': transaction['comment'],
            'type': transaction['type'],
            'lines': [{'accountId': line['account'],
                       'amount': line['amount'],
                       'counterpartyId': line['counterparty'],
                       'tags': line['tags']
                       }
                      for line in transaction['lines']
                      ]
        })

    def delete_transaction(self, transaction_id):
        self.__post('/transaction/%s/delete' % transaction_id)

    def delete_bmo_transactions(self, bmo_id):
        self.__post('/bmo-transactions/%s/delete' % bmo_id)

    def erase(self):
        self.__post('/erase')

    def enable_balance_cache(self):
        self.__post('/enable-balance-cache')

    def disable_balance_cache(self):
        self.__post('/disable-balance-cache')

    def take_snapshot(self, snapshot_time):
        logger.info('taking snapshot for %s with snapshot_time=%s' % (self.company_id, snapshot_time))
        if snapshot_time:
            self.__post('/snapshot/create?date=%s' % snapshot_time)
        else:
            self.__post('/snapshot/create')

    @staticmethod
    def __accountifie_svc_url():
        return api_func('environment', 'variable', 'ACCOUNTIFIE_SVC_URL')


"""
Adapted with permission from ReportLab's DocEngine framework
"""



import os

from django import forms
from django.conf import settings
from django.contrib.auth.forms import AuthenticationForm, PasswordChangeForm, \
        PasswordResetForm, SetPasswordForm

from accountifie.toolkit.forms import BootstrapMixin


class CommonAuthenticationForm(BootstrapMixin, AuthenticationForm):
    """
    adding attrs towards bootstrap ui
    """
    bs_horizontal = False

    def __init__(self, request=None, *args, **kwargs):
        super(CommonAuthenticationForm, self).__init__(request=request, *args, **kwargs)

        #apps with self registration are email only.

        if getattr(settings, 'DOCENGINE_REGISTRATION', False) and getattr(settings, 'REGISTRATION_OPEN', False):
            _user_label = _user_placeholder = "Email"
        else:
            _user_label = 'ID'
            _user_placeholder = 'Email or Username'
        _usr_kls = self.fields['username'].widget.attrs.get('class','')
        if len(_usr_kls)>0:
            _usr_kls+=' '
        _passwd_kls = self.fields['password'].widget.attrs.get('class','')
        if len(_passwd_kls)>0:
            _passwd_kls+=' '
        self.fields['username'].label = _user_label
        self.fields['username'].widget.attrs.update({'class': _usr_kls+'input-block-level',
                                                     'placeholder': _user_placeholder})
        self.fields['password'].widget.attrs.update({'class':_usr_kls+'input-block-level', 
                                                     'placeholder':'Password '})


class MultiCommonAuthenticationForm(CommonAuthenticationForm):

    subdomain = forms.CharField(max_length=100, widget=forms.HiddenInput(), required=False)

    def __init__(self, request=None, *args, **kwargs):
        super(MultiCommonAuthenticationForm, self).__init__(request=request, *args, **kwargs)
        _referer = request.META.get('HTTP_REFERER', '')
        if _referer:
            _referer = '/'.join(_referer.split('/', 3))
        if request and hasattr(request, 'session') and 'subdomain' in request.session:
            self.fields['subdomain'].initial = request.session['subdomain']
        elif _referer not in settings.LOGIN_URL:
            self.fields['subdomain'].initial=_referer.split('/')[-1]
        elif os.environ.get('DOMAIN', False):
            self.fields['subdomain'].initial = os.environ['DOMAIN']


class CommonPasswordChangeForm(BootstrapMixin, PasswordChangeForm):

    pass


class CommonPasswordResetForm(BootstrapMixin, PasswordResetForm):

    pass


class CommonSetPasswordForm(BootstrapMixin, SetPasswordForm):

    pass




# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('environment', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='variable',
            name='value',
            field=models.CharField(max_length=100),
        ),
    ]



"""
Strategy that performs transaction querying locally.
The tried and tested way of doing things, though not as performant as
query_manager_remote_strategy.py.
See query_manager_strategy.py for interface docs.
Use query_manager_strategy_factory.py to get an instance of this class.
"""
from django.db.models import Sum

from accountifie.gl.models import TranLine

from .query_manager_strategy import QueryManagerStrategy


class QueryManagerPostgresStrategy(QueryManagerStrategy):

    def cp_balances_for_dates(self, company_id, account_ids, dates):
        bals = {}
        for dt in dates:
            qs = TranLine.objects \
                         .filter(company_id=company_id) \
                         .filter(date__lte=dt)
            if account_ids:
                qs = qs.filter(account_id__in=account_ids)

            rslts = qs.values('counterparty_id') \
                      .annotate(Sum('amount'))
            bals[dt] = dict((l['counterparty_id'], l['amount__sum']) for l in rslts)

        return bals

    def account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra,filter_closing_entries=False):
        bals = {}
        for dt in dates:
            qs = TranLine.objects \
                         .filter(company_id=company_id) \
                         .filter(date__lte=dt)
            if filter_closing_entries:
                qs = qs.exclude(closing_entry=True)
            if account_ids:
                qs = qs.filter(account_id__in=account_ids)
            if with_counterparties:
                qs = qs.filter(counterparty_id__in=with_counterparties)

            rslts = qs.values('account_id') \
                      .annotate(Sum('amount'))
            bals[dt] = dict((l['account_id'], l['amount__sum']) for l in rslts)

        return bals



"""
Adapted with permission from ReportLab's DocEngine framework
"""



import os

from datetime import timedelta
from decimal import Decimal
from collections import namedtuple
from subprocess import Popen, PIPE
from django.conf import settings
from importlib import import_module


def run_shell_command(command, cwd):
    """
    Run command in shell and return results.
    """
    p = Popen(command, shell=True, cwd=cwd, stdout=PIPE, stderr=PIPE)
    stdout = p.communicate()[0]
    if stdout:
        stdout = stdout.strip()
    return stdout

_ntuple_diskusage = namedtuple('usage', 'total used free percentage')


def sizeof_fmt(num):
    for x in ['bytes','KB','MB','GB','TB']:
        if num < 1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0


# as per http://stackoverflow.com/questions/4274899/get-actual-disk-space
def disk_usage(path):
    
    """Return disk usage statistics about the given path.

    Returned valus is a named tuple with attributes 'total', 'used' and
    'free', which are the amount of total, used and free space, in bytes.
    """
    st = os.statvfs(path)
    free = st.f_bavail * st.f_frsize
    total = st.f_blocks * st.f_frsize
    used = (st.f_blocks - st.f_bfree) * st.f_frsize
    percentage = "%.2f" % (float(used)/float(total)*100) # make sure displays naturally with 2 decimal places
    percentage = Decimal(percentage)
    return _ntuple_diskusage(sizeof_fmt(total), sizeof_fmt(used), sizeof_fmt(free), percentage)


def server_uptime():
    uptimeString = 'unknown'
    with open('/proc/uptime', 'r') as f:
        uptime_seconds = float(f.readline().split()[0])
        uptime_string = str(timedelta(seconds = uptime_seconds))
    return uptime_string
    

def folder_size(folder):
    total_size = os.path.getsize(folder)
    for item in os.listdir(folder):
        itempath = os.path.join(folder, item)
        if os.path.isfile(itempath):
            total_size += os.path.getsize(itempath)
        elif os.path.isdir(itempath):
            total_size += folder_size(itempath)
    return total_size


def loggers():
    LOGGING_APPS_START = ('docengine', os.path.basename(settings.PROJECT_DIR) )
    

def get_available_stats():
    "Finds apps with a stats.py, gets back the dictionary contents"
    found = []
    for app_name in settings.INSTALLED_APPS:
        try:
            stats = import_module('.stats', app_name)
            if hasattr(stats, 'SQL_QUERIES'):
                for (stat_name, help_text, query) in stats.SQL_QUERIES:
                    found.append(
                        dict(app_name=app_name, 
                            stat_name=stat_name, 
                            help_text=help_text, 
                            query=query)
                        )
        except ImportError:
            continue
    return found

def get_configuration_warnings():
    "Checks a bunch of things which were not in docengine earlier but are now"
    warnings = []


    for (filename, caution) in [
        ('project/static/css/base.css', 
            """Allows you to override the CSS for your site, including templates
            provided by docengine itself.  An empty file will do."""),
        ('project/templates/base.html', 
            """Allows you to determine the HTML template for your site.  Docengine
            will expect to find this somewhere.  The minimum it needs to contain is:  {% extends 'common/base.html' %}"""),

        ]:
        expected_file_name = os.path.join(settings.ENVIRON_DIR, filename)
        if not os.path.isfile(expected_file_name):
            warnings.append(("Missing file", filename, caution))
    return warnings







# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-12-16 23:04
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0004_auto_20171126_2010'),
    ]

    operations = [
        migrations.AlterModelTable(
            name='address',
            table='common_address',
        ),
        migrations.AlterModelTable(
            name='log',
            table='common_log',
        ),
        migrations.AlterModelTable(
            name='taskmonitor',
            table='common_taskmonitor',
        ),
    ]



"""
Adapted with permission from ReportLab's DocEngine framework
"""




# Create your views here.
import os, sys
import hashlib
import logging
import json
from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import login_required
from django.contrib.auth.views import login as auth_login

from django.contrib.sites.shortcuts import get_current_site

from django.db.models import Q
from django.http import HttpResponse, Http404, HttpResponseRedirect, HttpResponseServerError, \
        HttpResponseNotFound
from django.shortcuts import render, get_object_or_404, resolve_url
from django.template import RequestContext, loader
from django.template.response import TemplateResponse
from django.views.decorators.debug import sensitive_post_parameters
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect
from django.utils.http import is_safe_url



from .models import TaskMonitor
from .forms import CommonAuthenticationForm, MultiCommonAuthenticationForm


#currently, this matches our logging config and anything called 'default' goes in the database
#this doesn't sound right but it's a start.
logger = logging.getLogger('default')


# HTTP Error 500
def custom_500(request):
    type, value, tb = sys.exc_info(),

    response = render(request, '500.html', {'message': value})
    response.status_code = 200
    return response


def base_templates(request):
    '''
    To be used as a context_processor so as to provide a fallback if the hosting
    project does not have either base.html or base_site.html templates
    '''
    _docengine_registration = getattr(settings, 'DOCENGINE_REGISTRATION', False)
    _registration_open = getattr(settings, 'REGISTRATION_OPEN', True)
    _USER_PROFILE = getattr(settings, 'REGISTER_USER_PROFILE', getattr(settings,
                                                                       'DOCENGINE_USER_PROFILE',
                                                                      None))
    _base_prefix=''
    if hasattr(request, 'session') and 'subdomain' in request.session:
        if hasattr(settings, 'DOMAINS_PATHS'):
            _base_prefix = dict(settings.DOMAINS_PATHS).get(request.session['subdomain'], '')
        else:
            _base_prefix = request.session['subdomain'] + '/'

    context = {
            'client_project': getattr(settings, 'CLIENT_PROJECT', 'DocEngine'),
            'base_template': loader.select_template([_base_prefix+'base.html', 'base.html', 'common/base.html']),
            'base_site_template': loader.select_template([_base_prefix+'base_site.html', 'base_site.html', 'common/base_site.html']),
            'REGISTRATION_OPEN': _docengine_registration and _registration_open,
            'HAS_PROFILE': isinstance(_USER_PROFILE, str),
            'PROJECT_TITLE': os.path.split(settings.ENVIRON_DIR)[1]
        }

    return context


@sensitive_post_parameters()
@csrf_protect
@never_cache
def login(request, template_name='common/login.html',
          redirect_field_name=REDIRECT_FIELD_NAME,
          authentication_form=CommonAuthenticationForm,
          current_app=None, extra_context=None):
    """
    Displays the login form and handles the login action.
    """
    if 'accountifie.middleware.ssl.SSLRedirect' not in settings.MIDDLEWARE_CLASSES and\
       'docengine.common.ssl.SecureRequiredMiddleware' not in \
                            settings.MIDDLEWARE_CLASSES and not getattr(settings,
                                                                    'DOCENGINE_HTTPS_SUPPORT',
                                                                        False):

        from django.core.exceptions import ImproperlyConfigured
        msg = 'Authentication is required to run under HTTPS -- Please append\
      "docengine.common.ssl.SSLRedirect" or "docengine.common.ssl.SecureRequiredMiddleware" to MIDDLEWARE_CLASSES in project settings'
        raise ImproperlyConfigured(msg)
    _login_indirect = False
    # the LOGIN_URL refers a absolute URI
    if settings.LOGIN_URL.startswith('http'):
        _login_indirect = True
        _login_server = ('/').join(settings.LOGIN_URL.split('/')[:3])
        if request.get_host() not in settings.LOGIN_URL:
            # get the contents up to the 3rd slash
            return HttpResponseRedirect(_login_server+request.get_full_path())

    if _login_indirect:
        authentication_form = MultiCommonAuthenticationForm

    if request.method == "POST":
        redirect_to = request.POST.get(redirect_field_name, '')

        form = authentication_form(request, data=request.POST)
        if form.is_valid():

            # Ensure the user-originating redirection url is safe.
            if not is_safe_url(url=redirect_to, host=request.get_host()):
                redirect_to = resolve_url(settings.LOGIN_REDIRECT_URL)

            # Okay, security check complete. Log the user in.
            auth_login(request, form.get_user())
            if _login_indirect:
                request.session['subdomain'] = form.cleaned_data['subdomain']
            return HttpResponseRedirect(redirect_to)
    else:
        redirect_to = request.GET.get(redirect_field_name, '')
        request.session.set_test_cookie()
        form = authentication_form(request)

    current_site = get_current_site(request)

    context = {
        'form': form,
        redirect_field_name: redirect_to,
        'site': current_site,
        'site_name': current_site.name,
    }
    if extra_context is not None:
        context.update(extra_context)
    if current_app:
        request.current_app = current_app
    return TemplateResponse(request, template_name, context)


def processes(request):
    message = ''
    sortField = 'memory'
    if request.POST:
            kill = request.POST.get('kill', '')
            if kill:
                message = '<p style="color: red;">Process %s was killed. Message returned: %s</p>' % (kill, os.popen('kill %s' % kill).read())
            sortField = request.POST.get('sortField', 'memory')
    processes = []
    fields = 'pid,ppid,user,time,state,%cpu,vsz,command'
    ps = os.popen('ps -o%s -w -w -w -ax | grep python' % fields).read().splitlines()
    ps = [Process(p) for p in ps if p.strip()]

    ppid = str(os.getppid())
    if ppid=='1': parent = pid
    else:
            pid = str(os.getpid())
            p = [p for p in ps if p.pid==pid]
            if p:
                    parent = p[0].ppid

    fields = fields.split(',')
    for p in ps:
            if p.user != 'rptlab' or 'grep' in p.command:
                    continue
            if p.pid==parent:
                    p.color = 'green'
            elif parent!='UNKNOWN' and p.ppid!=parent:
                    continue
            processes.append(p)
    processes = sortByField(processes, sortField)
    params = {'user': request.user,
              'title': 'Diagnose Processes',
              'host': request.META.get('HTTP_HOST', 'Unknown'),
              'processes': processes,
              'parent': parent,
              'message': message,
              'sortField': sortField}
    return render(request, 'common/processes.html', params)

def index(request):
    # Used on first setup, until you define a project-specific one which overrides it
    context = RequestContext(request)
    return render(request, 'common/index.html', context)


def test7491(request):
    # Used by alertra to check servers are working
    delay = request.REQUEST.get('delay', None)
    if delay is not None: time.sleep(float(delay))
    return HttpResponse('I am alive!', content_type='text/plain; charset=utf-8')

@login_required
def deliberateError(*args,**kwds):
    raise ValueError('Deliberately raised error')

class Process(object):

    fields = ['pid', 'ppid', 'user', 'time', 'state', '%cpu', 'vsz', 'command']
    names = [('pid', 'PID'), ('ppid', 'PPID'), ('user', 'User'), ('time', 'Time'), ('state', 'State'), ('%cpu', 'CPU%'), ('vsz', 'Size'), ('command', 'Command')]
    color = 'black'

    def __init__(self, s):
        self.s = s
        s = s.split()
        s[len(self.fields) - 1] = ' '.join(s[len(self.fields) - 1:])
        s = s[:len(self.fields)]
        for i, field in enumerate(self.fields):
            setattr(self, field, s[i].strip().lstrip())
        self.memory = int(self.vsz)
        if self.state == 'Z' or int(self.memory) > 200000:
            self.color = 'red'
        if len(str(self.vsz)) >= 6:
            self.vsz = '%sM' % str(self.memory / 1024)
        else:
            self.vsz = '%sK' % self.vsz

    def __unicode__(self):
        return force_unicode(self.s)

def sortByField(l, field):
    def fieldSort(a, b):
        if getattr(a, field, '') > getattr(b, field, ''): return 1
        if getattr(a, field, '') == getattr(b, field, ''): return 0
        return -1
    l.sort(fieldSort)
    return l

def media_types(app_name=''):
    '''
    Lists folders in media directory
    Takes optional app_name for which user folder to look into

    '''
    media_path = os.path.join(setting.MEDIA_ROOT, app_name)
    try:
        return [entry for entry in os.listdir(media_path) if os.path.isdir(os.path.join(media_path, entry)) ]
    except OSError:
        raise True("%s does not exist or it is not a folder" % media_path)


def media(media_type, app_name=''):
    '''
    Returns the list of [user] media items
    Takes optional app_name for which user folder to look into
    media_type = [ image | font | pallete | style | ... ]
    '''
    if media_type in media_type(app_name):
        media_path = os.path.join(settings.MEDIA_ROOT, app_name, media_type)
        if os.path.isdir(os.path.join(media_path)):
            return os.listdir(media_path)
    #no such media type
    return []


@login_required
def upload(request):
    "Handles user file uploads into a queue"

    result = {'success': False}
    if request.method == 'POST':
        if not os.path.isdir(settings.UPLOAD_DIR):
            os.makedirs(settings.UPLOAD_DIR)

        if len(request.FILES) == 1:
            upload = list(request.FILES.values())[0]
            rawdata = upload.read()
            result['bytes_received'] = len(rawdata)

            hasher = hashlib.md5()
            hasher.update(rawdata)
            ident = hasher.hexdigest()[0:6]

            info = {
                'filename': upload.name,
                'user': request.user.username,
                'remote_addr': request.META['REMOTE_ADDR']
                }
            datafilename = os.path.join(settings.UPLOAD_DIR, '%s.dat' % ident)
            infofilename = os.path.join(settings.UPLOAD_DIR, '%s.inf' % ident)
            open(datafilename, 'wb').write(rawdata)
            open(infofilename, 'wb').write(json.dumps(info))

            result['success'] = True
            result['message'] = 'File saved as "%s.dat"' % ident
            result['id'] = ident

            logger.info('User %s uploaded file %s, saved as %s' % (request.user, datafilename, ident))

            return HttpResponse(json.dumps(result), content_type='application/javascript; charset=utf8')
    else:
        result['message'] = 'You need to do a POST request to upload anything'
    return HttpResponse(json.dumps(result))


def server_error(request, template_name='500.html'):
    # like django.views.defaults.server_error, but uses RequestContext
    t = loader.select_template([template_name, 'common/500.html'])
    try:
        request_context = RequestContext(request, {})
    except:
        request_context = {}
    # will allow raising an exception if tempalte can't be rendered
    response = t.render(request_context)
    return HttpResponseServerError(response)


def not_found(request, template_name='404.html'):
    # like django.views.defaults.server_error, but uses RequestContext
    t = loader.select_template([template_name, 'common/404.html'])
    try:
        request_context = RequestContext(request, {})
    except:
        request_context = {}
    # will allow raising an exception if tempalte can't be rendered
    response = t.render(request_context)
    return HttpResponseNotFound(response)


@login_required
def check_all_tasks(request):
    '''endpoint to check up on tasks that are either still running
    or that have finished and were started less then 5 min ago'''
    tasks = []
    now = datetime.now()
    five_min = timedelta(seconds=300)
    past = now - five_min
    monitors = TaskMonitor.objects.filter(Q(creation_timestamp__gte=past) | Q(task_state='started'))
    for m in monitors:
        tasks.append(dict(task_id=m.task_id,
                          task_name=m.task_name,
                          traceback=m.traceback,
                          task_state=m.task_state,
                          percent_complete=m.percent_complete,
                          timestamp=m.creation_timestamp.strftime("%A, %d. %B %Y %I:%M%p"),
                          )
                     )
    return HttpResponse(json.dumps(tasks))



from django.contrib import admin

from .models import *


class LogAdmin(admin.ModelAdmin):
    list_display = ('level', 'time', 'message', 'corrId', 'traceback', 'request',)
    
admin.site.register(Log, LogAdmin)

class AddressAdmin(admin.ModelAdmin):
    list_display = ('label', 'name', 'company',)
    
admin.site.register(Address, AddressAdmin)



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Address',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=100, null=True, blank=True)),
                ('company', models.CharField(max_length=100, null=True, blank=True)),
                ('address1', models.CharField(max_length=100, null=True, blank=True)),
                ('address2', models.CharField(max_length=100, null=True, blank=True)),
                ('city', models.CharField(max_length=50, null=True, blank=True)),
                ('postal_code', models.CharField(max_length=20, null=True, blank=True)),
                ('province', models.CharField(max_length=30, null=True, blank=True)),
                ('country', models.CharField(max_length=30, null=True, blank=True)),
                ('phone', models.CharField(max_length=30, null=True, blank=True)),
                ('email', models.EmailField(max_length=254, null=True, blank=True)),
            ],
        ),
    ]



import factory
from .models import Company, Counterparty, Account, ExternalAccount

class CompanyFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Company

    id = 'TEST_COMPANY'
    name = 'Test Company'
    cmpy_type = 'ALO'


class AccountFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Account


class CounterpartyFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Counterparty

    id = 'TEST_CP'
    name = 'Test Counterparty'

class ExternalAccountFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ExternalAccount



from multipledispatch import dispatch

from accountifie.gl.models import Counterparty

def counterparties(qstring={}):
    data = list(Counterparty.objects.order_by('id').values())
    return data


@dispatch(dict)
def counterparty(qstring):
    data = list(Counterparty.objects.order_by('id').values())
    return data


@dispatch(str, dict)
def counterparty(cp_id, qstring):
    cp = Counterparty.objects.filter(id=cp_id).first()
    if cp is None:
        return None
    else:
        return dict((k,v) for k,v in cp.__dict__.items() if k in ['id','name'])



"""
Adapted with permission from ReportLab's DocEngine framework
"""

import os
import json
from io import StringIO

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.contrib import messages
from django.db import transaction
from django.apps import apps
from django.core.serializers.json import DjangoJSONEncoder

from accountifie.toolkit.utils import get_default_company, csv_to_modelattr, create_instance, \
                                      dirty_key, get_company, utcnow

from accountifie.toolkit.forms import FileForm


import logging

logger = logging.getLogger('default')


DATA_ROOT = getattr(settings, 'DATA_DIR', os.path.join(settings.ENVIRON_DIR, 'data'))
INCOMING_ROOT = os.path.join(DATA_ROOT, 'incoming')
PROCESSED_ROOT = os.path.join(DATA_ROOT, 'processed')


@login_required
def upload_file(request, config):
    file_type = config['file_type']

    context = {'file_type': file_type,}     
    form = FileForm(request.POST, request.FILES)

    if form.is_valid():
        upload = list(request.FILES.values())[0]
        file_name_with_timestamp = save_file(upload)
        company = get_company(request)

        model = config.pop('model')
        unique = config.pop('unique')
        name_cleaner = config.pop('name_cleaner')
        value_cleaner = config.pop('value_cleaner')
        exclude = config.pop('exclude')
        post_process = config.pop('post_process')

        result = process_incoming_file(model, unique, name_cleaner, value_cleaner, exclude, post_process,
                                                 file_type=config['file_type'],
                                                 file_name=file_name_with_timestamp,
                                                 company=company,
                                                 task_title="%s data for %s" % (file_type, company)
                                                 )

        output = dict((k, result.get(k)) for k in ['saved', 'key_errors', 'value_errors', 'found', 'data'])
        output['dupes'] = len(result.get('dups', {}))
        context = {'data': json.dumps(output, cls=DjangoJSONEncoder, indent=2)}
        context['title'] = 'File Upload Results'
        msg = 'Loaded %s. Checked %d -- %d new, %d dupes.' % (config['file_type'],
                                                              output['found'],
                                                              len(output['saved']),
                                                              output['dupes'])
        messages.info(request, msg)
        return render(request, 'api_display.html', context)
    else:
        context['error'] = 'form not in valid format'
        context['title'] = 'File Upload Results'
        return render(request, 'api_display.html', context)


def save_file(f):
    new_name = '%s_%s' % (utcnow().strftime('%Y%m%d%H%M%S'), f._name)
    if not os.path.exists(INCOMING_ROOT):
        os.makedirs(INCOMING_ROOT)
    full_path = os.path.join(INCOMING_ROOT, new_name)
    with open(full_path, 'wb') as destination:
        for chunk in f.chunks():
            destination.write(chunk)

    return new_name 


def process_incoming_file(model, unique, name_cleaner, value_cleaner, exclude, post_process, **config):
    out = StringIO()
    file_name = config.pop('file_name')
    context = {'file_name': file_name}
    file_type = config.pop('file_type')
    
    incoming_name = os.path.join(INCOMING_ROOT, file_name)
    
    with open(incoming_name, 'U') as f:

        data = json.loads(load_file(file=f, stdout=out, model=model, unique=unique, name_cleaner=name_cleaner,
                                    value_cleaner=value_cleaner, exclude=exclude, post_process=post_process, **config))

    context['found'] = sum([len(data[k]) for k in data])
    context['data'] = data['saved']
    if data['saved']:
        context['example'] = data['saved'][0]
    elif data['key_errors']:
        context['example'] = data['key_errors']
    else:
        context['example'] = dict(no_keys_found='no values found')
    for k in [x for x in list(data.keys()) if x not in list(context.keys())]:
        context[k] = data[k]
    
    post_processing(file_type)
    
    # move the file into 'processed' directory
    if not os.path.exists(PROCESSED_ROOT):
        os.makedirs(PROCESSED_ROOT)
    processed_name = os.path.join(PROCESSED_ROOT, file_name)
    os.rename(incoming_name, processed_name)
    return context


def post_processing(file_type):
    """
    For example matching counterparties based on historical patterns
    """
    return


def load_file(**config):
    file = config.pop('file', None)
    company = config.get('company', get_default_company())
    model = config.get('model')
    unique = config.get('unique')
    name_cleaner = config.get('name_cleaner')
    value_cleaner = config.get('value_cleaner')
    exclude = config.get('exclude', [])
    data = csv_to_modelattr(file, name_cleaner=clean_csv_fields, company=company)
    #with transaction.atomic():
    results = save_data(data, company, model, unique, name_cleaner, value_cleaner, exclude)
        
    return json.dumps(results, ensure_ascii=False)
    
                
def save_data(data, company, model, unique, name_cleaner, value_cleaner, exclude):
    if name_cleaner == None:
        name_cleaner = lambda name: name
    if value_cleaner == None:
        value_cleaner = lambda name, value: value
    if unique == None:
        unique = lambda instance: instance
    model = apps.get_model(*model.split('.'))

    dups = []
    saved = []
    value_errors = []

    for index, row in enumerate(data):
        dirty = dirty_key(row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner)
        
        if dirty:
            #if a key is wrong the whole file will fail anyway
            return dict(dups=dups, saved=saved, key_errors=dirty, value_errors=value_errors)
        with transaction.atomic():
            unique_instance = create_instance(row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner,
                                              unique=unique, company=company, exclude=exclude)
        if unique_instance:
            try:
                unique_instance.save()
            except Exception as e:
                print('exception right here', e)
                message = getattr(e, 'messages', [row])[0]
                value_errors.append((e.__class__.__name__, index, message))
                continue
            
            #need to deconstruct the model instance for later json serialization
            saved.append(dict([(k, v)for k,v in list(unique_instance.__dict__.items()) if k in [f.name for f in unique_instance._meta.fields]]))
        else:
            decoded = dict((k.decode('utf-8', 'ignore'),v.decode('utf-8', 'ignore')) for k,v in row.items())
            dups.append(decoded)
    
    return dict(dups=dups, saved=saved, key_errors=[], value_errors=value_errors)


def clean_csv_fields(name):
    return name.lower().replace(' ', '_').replace('-', '_').replace('check_or_slip_#', 'check_or_slip')
    


import pandas as pd
import numpy as np
import itertools
from . import query_manager_strategy_factory

from django.conf import settings

import accountifie.toolkit.utils as utils
from accountifie.common.api import api_func

import logging
logger = logging.getLogger('default')


def _get_acct_list(acct_list, paths):
        if acct_list:
            return acct_list
        else:
            if paths:
                accts = list(itertools.chain(*[api_func('gl', 'path_accounts', p) for p in paths]))
                return [x['id'] for x in accts]
            else:
                return [x['id'] for x in api_func('gl', 'path_accounts', '')]


class QueryManager:

    def __init__(self, gl_strategy=None):
        if gl_strategy:
            if isinstance(gl_strategy, str):
                strategy_inst = query_manager_strategy_factory.QueryManagerStrategyFactory().get(strategy=gl_strategy)
            else:
                strategy_inst = gl_strategy
        else:
            strategy_inst = query_manager_strategy_factory.QueryManagerStrategyFactory().get()
        self.gl_strategy = strategy_inst


    def acct_balances(self, company_id, as_of):
        return self.gl_strategy \
                   .account_balances_for_dates(company_id,
                                               None,
                                               [as_of],
                                               None, None, None)[as_of]

    ## REFERENCES: 0 internal, 2 external
    def path_drilldown(self, company_id, dates, path, excl_contra=None, excl_interco=None):
        paths = api_func('gl', 'child_paths', path)
        output = self.pd_path_balances(company_id, dates, paths, excl_contra=excl_contra, excl_interco=excl_interco)
        return output
    
    
    def cp_balances(self, company_id, dates, paths=None, acct_list=None, excl_contra=None, excl_interco=False, with_tags=None, excl_tags=None):

        accts = _get_acct_list(acct_list, paths)
        dates_dict = dict((dt, utils.get_dates_dict(dates[dt])) for dt in dates)
        balances = self.gl_strategy.cp_balances_for_dates(company_id, accts, dates_dict)

        # filter empties
        for acct in balances:
            for dt in balances[acct]:
                for col in ['openingBalance', 'closingBalance']:
                    balances[acct][dt][col] = [l for l in balances[acct][dt][col] if abs(float(l['total'])) > 0.0]
        return balances


    def pd_path_balances(self, company_id, dates, paths, filter_zeros=True, assets=False, excl_contra=None, excl_interco=False, with_tags=None, excl_tags=None):

        path_accts = dict((p, [x['id'] for x in api_func('gl', 'path_accounts', p)]) for p in paths)
        acct_list = list(itertools.chain(*[path_accts[p] for p in paths]))

        dates_dict = dict((dt, utils.get_dates_dict(dates[dt])) for dt in dates)
        date_indexed_account_balances = self.gl_strategy.account_balances_for_dates(company_id, acct_list, dates_dict, None, excl_interco, excl_contra, with_tags, excl_tags)

        data = {}
        for dt in date_indexed_account_balances:
            account_balances = date_indexed_account_balances[dt]

            tuples = []
            for path in paths:
                path_sum = 0
                for account_id in path_accts[path]:
                    if account_id in account_balances:
                        path_sum -= account_balances[account_id]
                tuples.append((path, path_sum))

            data[dt] = dict(tuples)

        output = pd.DataFrame(data)

        if filter_zeros:
            output['empty'] = output.apply(lambda row: np.all(row.values == 0.0), axis=1)
            output = output[output['empty']==False]
            output.drop('empty', axis=1, inplace=True)

        # adjust assets sign
        if assets:
            asset_factor = output.index.map(lambda x: -1.0 if x[:6] == 'assets' else 1.0)
            for col in output.columns:
                output[col] = output[col] * asset_factor
        return output

    def pd_acct_balances(self, company_id, dates, paths=None, acct_list=None, excl_contra=None, excl_interco=False, cp=None, with_tags=None, excl_tags=None):

        if not acct_list:
            if paths:
                accts = list(itertools.chain(*[api_func('gl', 'path_accounts', p) for p in paths]))
                acct_list = [x['id'] for x in accts]
            else:
                acct_list = [x['id'] for x in api_func('gl', 'path_accounts', '')]

        dates_dict = dict((dt, utils.get_dates_dict(dates[dt])) for dt in dates)

        with_counterparties = [cp.id] if cp else None
        date_indexed_account_balances = self.gl_strategy.account_balances_for_dates(company_id, acct_list, dates_dict, with_counterparties, excl_interco, excl_contra, with_tags, excl_tags)

        # filter empties
        for dt in date_indexed_account_balances:
            d = date_indexed_account_balances[dt]
            date_indexed_account_balances[dt] = {k:d[k] for k in d if d[k] != 0.0}

        output = pd.DataFrame(date_indexed_account_balances).fillna(0)

        if paths:
            a = [[x['id'] for x in api_func('gl', 'path_accounts', path)] for path in paths]
            acct_list = list(itertools.chain(*a))
            return output[output.index.isin(acct_list)]
        elif acct_list:
            filtered_output = output[output.index.isin(acct_list)]
            return filtered_output

        return output

    def transaction_info(self, company_id, trans_id):
        return self.gl_strategy.get_transaction(company_id, trans_id)


    def transactions(self, company_id, from_date=settings.DATE_EARLY, to_date=settings.DATE_LATE):
        acct_list = [x['id'] for x in api_func('gl', 'account')]
        all_entries = self.gl_strategy.transactions(company_id, acct_list, from_date, to_date, 'end-of-month', None, None, None)

        return all_entries


    def pd_history(self, company_id, q_type, id, from_date=settings.DATE_EARLY, to_date=settings.DATE_LATE, excl_interco=None, excl_contra=None, incl=None, cp=None):
        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            excl_interco = True

        if q_type == 'account':
            acct_list = [id]
        elif q_type == 'account_list':
            acct_list = incl
        elif q_type=='path':
            acct_list = [x['id'] for x in api_func('gl', 'path_accounts', id, qstring={'excl': excl_contra, 'incl': incl})]
        else:
            raise ValueError('History not implemented for this type')

        with_counterparties = [cp] if cp else None

        all_entries = self.gl_strategy.transactions(company_id, acct_list, from_date, to_date, 'end-of-month', with_counterparties, excl_interco, excl_contra)
        
        if all_entries is None or len(all_entries) == 0:
            return pd.DataFrame()
    
        all_entries_pd = pd.DataFrame(all_entries)
        all_entries_pd.sort(['date', 'id'], ascending=[1, 0], inplace=True)
    
        cols = ['date', 'id', 'comment', 'account_id', 'contra_accts', 'counterparty', 'amount']
    
        top_line = pd.Series(dict((col, None) for col in cols))

        top_line['date'] = utils.day_before(all_entries_pd.iloc[0]['date'])
        top_line['amount'] = 0.0
        top_line['comment'] = 'Opening Balance'

        
        all_entries_pd.reset_index(drop=True, inplace=True)
        all_entries_pd.loc['starting'] = top_line
        all_entries_pd.sort_index(by='date', axis=0, inplace=True)
        all_entries_pd['balance'] = all_entries_pd['amount'].cumsum()
    
        return all_entries_pd[cols + ['balance']]

    def balance_by_cparty(self, company_id, acct_ids, from_date=settings.DATE_EARLY, to_date=settings.DATE_LATE):

        all_entries = self.gl_strategy.transactions(company_id, acct_ids, from_date, to_date, 'end-of-month', None, None, None)
        if all_entries is None or len(all_entries) == 0:
            return pd.DataFrame()

        entries_tbl = pd.DataFrame(all_entries).sort_index(by='date')
        cp_tbl = entries_tbl[['counterparty','amount']].groupby('counterparty').sum()

        filter_small = lambda x: x>0.99 or x<-0.99
        cp_tbl = cp_tbl[cp_tbl['amount'].map(filter_small)]

        cp_tbl = cp_tbl.sort_index(by='amount', ascending=False)
        return cp_tbl['amount']



"""Calendar module as recommended by Celko.

We have pre-joined database records for Year, Quarter,
Month and Day objects.  This means that if one needs
to total sales by quarter, or to make screens to
browse through VAT quarters or statements, it can be
done very efficiently in SQL.

MySQL date functions are good enough to group by end-of-month,
but other entities in an accounts system can often be of a
monthly or quarterly nature, so find it easy to have a
month_id or quarter_id.

Years are identified as '2009', quarters as '2009Q1'
and months as '2009M07' to ensure alpha order equals chrono
order.

The populate function can be called on startup to ensure
it is always populated for the current year; and also
from
   manage.py populate_calendar --from=YYYYMMDD --to=YYYYMMDD





"""
import re
from datetime import date, timedelta, datetime
from django.db import transaction

#for quick checking if we have a valid period id.  Does not rule out e.g. week 59 or month 19;-)
PAT_YEAR = re.compile("^\d{4}$")
PAT_MONTH = re.compile("^\d{4}M(0[1-9]{1}|10|11|12)$")
PAT_QUARTER = re.compile("^\d{4}Q[0-4]$")
PAT_HALF = re.compile("^\d{4}H[1-2]$")
PAT_WEEK = re.compile("^\d{4}W(0[1-9]|[1-4][0-9]|5[0-3]$)")
PAT_DAY = re.compile("^(D)(\d{4}-\d{1,2}-\d{1,2})")

#combine above 
PAT_PERIOD_ID = re.compile("|".join([PAT_YEAR.pattern, PAT_MONTH.pattern, PAT_QUARTER.pattern, PAT_HALF.pattern, PAT_WEEK.pattern, PAT_DAY.pattern]))


@transaction.atomic
def populate(from_date=None,
             to_date=None,
             verbose=False):
    """Ensure calendar has all dates in the range.
    """
    if not from_date:
        now = date.today()
        from_date = date(now.year, 1, 1)
    if not to_date:
        now = date.today()
        next_new_year = date(now.year + 1, 1, 1)
        to_date = next_new_year - timedelta(1)


    from .models import Day, Month, Quarter, Year, Week
    
    #quick check to make this cheap to run on startup.  
    days = (to_date - from_date).days + 1
    already = Day.objects.filter(id__gte=from_date, id__lte=to_date).count()
    if days == already:
        print("Already populated")
        return

    #get all the records for efficiency so we can create
    #anything missing

    days = dict((day.id, day) for day in Day.objects.all())
    months = dict((month.id, month) for month in Month.objects.all())
    quarters = dict((qtr.id, qtr) for qtr in Quarter.objects.all())
    years = dict((year.id, year) for year in Year.objects.all())
    
    
    if verbose:
        print('populating calendar from %s to %s' % (from_date, to_date))
    d = from_date
    while d <= to_date:
        year_id = d.year
        quarter_num = ((d.month-1)/3)+1
        quarter_id = '%04dQ%01d' % (d.year, quarter_num)
        month_id = '%04dM%02d' % (d.year, d.month)

        #find or create the Year object
        if year_id not in years:
            djyear = Year(id=d.year)
            djyear.save()
            years[d.year] = djyear
            if verbose: print('    added year %s' % d.year)
        else:
            djyear = years[year_id]
            
        #find or create the Quarter object
        if quarter_id not in quarters:
            djquarter = Quarter(id=quarter_id, year=djyear)
            djquarter.save()
            quarters[quarter_id] = djquarter
            if verbose: print('        added quarter %s' % quarter_id)
        else:
            djquarter = quarters[quarter_id]
        
        #find or create the Month object
        if month_id not in months:
            djmonth = Month(id=month_id,
                            year=djyear,
                            quarter=djquarter)
            djmonth.save()
            months[month_id] = djmonth
            if verbose: print('            added month %s' % month_id)
        else:
            djmonth = months[month_id]
        
        #find or create the Day object
        if d not in days:
            djday = Day(id=d,
                        year=djyear,
                        quarter=djquarter,
                        month=djmonth)
            djday.save()
            days[d] = djday
            if verbose: print('              added day %s' % d)
        
        d = d + timedelta(1)


    #now for the weeks
    d = from_date
    #roll backwards to monday, in case of an incomplete one from last year
    while d.weekday() != 0:
        d = d - timedelta(days=1)

    while d <= to_date:
        yyyy, wk, wkday = d.isocalendar()
        week_id = '%04dW%02d' % (yyyy, wk)
        try:
            week = Week.objects.get(id=week_id)
        except Week.DoesNotExist:
            week = Week(id=week_id)
            week.year_id = yyyy
            week.first_day = d
            week.last_day = d + timedelta(days=6)
            #join up all the joins.
            
            try:
                week.start_month = Month.objects.get(id="%04dM%02d" % (yyyy, d.month))
                week.end_month = Month.objects.get(id="%04dM%02d" % (yyyy, d.month))
                week.save()
                if verbose: print(("created week %s" % week_id))
            except Month.DoesNotExist:
                #hit year end, don't bother saving this week
                if verbose: print(("No related month, skipping %s" % week_id))
                
        d = d + timedelta(days=7)



    return



def start_of_month(d):
    return date(d.year, d.month, 1)

def end_of_month(d):
    year = d.year
    month = d.month + 1
    if month > 12:
        year += 1
        month = 1
    return date(year, month, 1) - timedelta(days=1)

def end_of_previous_month(d):
    return start_of_month(d) - timedelta(days=1)


#brute force depending on month index, gives month-year for each input month
_QUARTER_RANGES = [
    (None,None),  #no month zero
    (1,1,3,31),
    (1,1,3,31), 
    (1,1,3,31), 
    (4,1,6,30),
    (4,1,6,30), 
    (4,1,6,30), 
    (7,1,9,30),
    (7,1,9,30), 
    (7,1,9,30), 
    (10,1,12,31),
    (10,1,12,31),   
    (10,1,12,31),   
    ]   

def start_of_quarter(dt):
    a,b,c,d = _QUARTER_RANGES[dt.month]
    return date(dt.year, a, b)

def end_of_quarter(dt):
    a,b,c,d = _QUARTER_RANGES[dt.month]
    return date(dt.year, c, d)

def start_of_half(dt):
    if dt.month < 7:
        return date(dt.year, 1, 1)
    else:
        return date(dt.year, 6, 1)

def end_of_half(dt):
    if dt.month < 7:
        return date(dt.year, 6, 30)
    else:
        return date(dt.year, 12, 31)


def start_of_year(dt):  
    return date(dt.year, 1, 1)

def end_of_year(dt):    
    return date(dt.year, 12, 31)

def start_of_week(dt):
    year, week, weekday = dt.isocalendar()
    return dt - timedelta(days=weekday-1)

def end_of_week(dt):
    year, week, weekday = dt.isocalendar()
    return dt + timedelta(days=(7-weekday))


def is_period_id(text):
    if not text:
        return False

    match = PAT_PERIOD_ID.match(str(text))
    return (match is not None)


def is_iso_date(text):
    try:
        return datetime.strptime(text, '%Y-%M-%d')
    except ValueError:
        return False

def parse_iso_date(text):
    if isinstance(text, datetime.date):
        return text
    return datetime.strptime(text, '%Y-%M-%d')

def this_quarter_id():
    "Returns quarter identifier for present"
    t = date.today()
    q = int((t.month + 1)/3)
    return '%04dQ%d' % (t.year, q)

def this_half_id():
    "Returns half identifier for present"
    t = date.today()
    if t.month < 7:
        half = 1
    else:
        half = 2
    return '%04dH%d' % (t.year, half)

def this_month_id():
    "Returns month identifier for present"
    t = date.today()
    return '%04dM%02d' % (t.year, t.month)
    

def this_week_id():
    "Returns week identifier for present"
    y, w, wd = date.today().isocalendar()
    return '%04dW%02d' % (y, w)
    

def tofirstdayinisoweek(year, week):
    #from stackoverflow
    ret = strptime('%04d-%02d-1' % (year, week), '%Y-%W-%w')
    if date(year, 1, 4).isoweekday() > 4:
        ret -= timedelta(days=7)
    return ret

def previous_period(id):
    "Returns identifier for preceding period"
    year = int(id[0:4])
    part2 = int(id[5:])  #month, week, quarter
    if id[4] == 'W':
        import isoweek
        monday = isoweek.Week(year, part2).monday()
        previous_monday = monday - timedelta(days=7)
        y,w,wd = previous_monday.isocalendar()
        return '%04dW%02d' % (y, w)
    elif id[4] == 'M':
        part2 = part2 - 1
        if part2 == 0:
            part2 = 12
            year -= 1
        return '%04dM%02d' % (year, part2)
    elif id[4] == 'Q':
        part2 = part2 - 1
        if part2 == 0:
            part2 = 4
            year -= 1
        return '%04dQ%01d' % (year, part2)
    elif id[4] == 'H':
        if part2 == 1:
            year -= 1
            half = 2
        else:
            half = 1
        return '%04dH%01d' % (year, half)
    else:
        raise ValueError("Unexpected date identifier %s" % id)


def next_period(id):
    "Returns identifier for next period"
    year = int(id[0:4])
    part2 = int(id[5:])  #month, week, quarter
    if id[4] == 'W':        
        import isoweek
        monday = isoweek.Week(year, part2).monday()
        next_monday = monday + timedelta(days=7)
        y,w,wd = next_monday.isocalendar()
        return '%04dW%02d' % (y, w)
    elif id[4] == 'M':
        part2 = part2 + 1
        if part2 == 13:
            part2 = 1
            year += 1
        return '%04dM%02d' % (year, part2)
    elif id[4] == 'Q':
        part2 = part2 + 1
        if part2 == 5:
            part2 = 1
            year += 1
        return '%04dQ%01d' % (year, part2)
    elif id[4] == 'H':
        if part2 == 2:
            year += 1
            half = 1
        else:
            half = 2
        return '%04dH%01d' % (year, half)

    else:
        raise ValueError("Unexpected date identifier %s" % id)


def start_of_period(period_id):
    pid = period_id
    assert is_period_id(pid)
    year = int(pid[0:4])
    if len(pid) > 5: 
        part2 = int(pid[5:])  #month, week, quarter
        sep = pid[4]
    else:
        part2 = None
        sep = 'Y'
    if sep == 'W':        
        import isoweek
        return isoweek.Week(year, part2).monday()
    elif sep == 'M':
        return date(year, part2, 1)
    elif sep == 'Q':
        month_id = (part2 - 1) * 3 + 1
        return date(year, month_id, 1)
    elif sep == 'H':
        return start_of_half(date(year, (part2*6), 1))
    elif sep == 'Y':
        return date(year, 1, 1)
    else:
        raise ValueError("Unexpected date identifier %s" % id)
    
def end_of_period(period_id):
    assert is_period_id(period_id)
    pid = period_id
    year = int(period_id[0:4])
    if len(pid) > 5: 
        part2 = int(pid[5:])  #month, week, quarter
        sep = pid[4]
    else:
        part2 = None
        sep = 'Y'
    if sep == 'W':        
        import isoweek
        return isoweek.Week(year, part2).sunday()
    elif sep == 'M':
        return end_of_month(date(year, part2, 1))
    elif sep == 'Q':
        return end_of_quarter(date(year, (part2*3), 1))
    elif sep == 'H':
        return end_of_half(date(year, (part2*6), 1))
    elif sep == 'Y':
        return date(year, 12, 31)
    else:
        raise ValueError("Unexpected date identifier %s" % id)





def supply_date_context():
    """Supply a dictionary with stuff for time navigation in templates

    Contains string IDs for everything so you can build URLs.
    """
    #array we can index into with 1..12
    MONTHNAMES = [None] + "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec".split()
    d = {}
    today = date.today()
    d['today'] = today.isoformat()
    d['tomorrow'] = (today + timedelta(days=1)).isoformat()
    d['yesterday'] = (today - timedelta(days=1)).isoformat()

    d['this_year'] = str(today.year)
    d['last_year'] = str(today.year - 1)
    d['next_year'] = str(today.year + 1)

    d['this_quarter'] = this_quarter_id()
    d['next_quarter'] = next_period(d['this_quarter'])
    d['last_quarter'] = previous_period(d['this_quarter'])
    d['two_quarters_ago'] = previous_period(d['last_quarter'])

    d['this_month'] = this_month_id()
    d['next_month'] = next_period(d['this_month'])
    d['last_month'] = previous_period(d['this_month'])
    d['two_months_ago'] = previous_period(d['last_month'])

    d['this_month_mmm'] = MONTHNAMES[today.month]
    last_month = end_of_previous_month(today)
    d['last_month_mmm'] = MONTHNAMES[last_month.month]
    two_months_ago = end_of_previous_month(last_month)
    d['two_months_ago_mmm'] = MONTHNAMES[two_months_ago.month]

    d['this_week'] = this_week_id()
    d['next_week'] = next_period(d['this_week'])
    d['last_week'] = previous_period(d['this_week'])
    d['two_weeks_ago'] = previous_period(d['last_week'])

    return d


if __name__=='__main__':
    populate()



import datetime

from djcelery.models import TaskMeta

from django.utils.safestring import mark_safe

_STATUS_COLOR_CLASS = dict(list(zip(('IN_PROGRESS', 'COMPLETED', 'FAILED', 'UNKNOWN') , ('info', 'success', 'danger', 'warning'))))

def colorise(status):
    return _STATUS_COLOR_CLASS[status]

  
def status_tag(status):
	return mark_safe('<span class="label label-%s">%s</span>' % (
                colorise(status), status)
    )

def download_link(url):
	return mark_safe('<a href="%s">Download results</a>' % url
    )

def _get_info(t):
	rslt = t.get('result', {})
	output = rslt.get('output')

	data = {}
	data['task_id'] = t['task_id']
	data['date_done'] = t['date_done']
	
	if rslt:
		data['task_status'] = rslt.get('status', 'UNKNOWN')
		data['task_name'] = rslt.get('task_name')

		if output:
			data['return_value'] = output.get('return_value')
			download = output.get('download')
			if download:
				data['download'] = download_link(output.get('download'))
			data['error'] = output.get('error')
		
	return data

def tasks_in_progress(qstring={}):
	unfinished = list(TaskMeta.objects.exclude(status='SUCCESS').values())
	return [_get_info(t) for t in unfinished if t['result']['status']=='IN_PROGRESS']


def tasks_just_finished(qstring={}):
	RECENT = 60 * 60 * 24 * 7 # last week
	now = datetime.datetime.now()
	cutoff = now - datetime.timedelta(seconds=RECENT)
	data = list(TaskMeta.objects.filter(date_done__gte=cutoff) \
						   .order_by('-date_done').values())
	return [_get_info(t) for t in data]



"""
Based on Andy Robinson's DoubleTalk accounting framework; used
with permission
"""

from decimal import Decimal
import logging
from deepdiff import DeepDiff

from django.db import transaction
from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist

import accountifie.query.query_manager_strategy_factory as QMSF
from accountifie.environment.models import Variable
from accountifie.gl.models import TranLine, Company, Account, Counterparty

DZERO = Decimal('0')

logger = logging.getLogger('default')
GL_STRATEGY = 'postgres'

def _model_to_id(x):
    return x if type(x) == str else x.id


def get_gl_strategy():
    dflt_strategy = Variable.objects \
                            .filter(key='DEFAULT_GL_STRATEGY') \
                            .first()
    return dflt_strategy.value if dflt_strategy else GL_STRATEGY


def recalc_all():
    for model in apps.get_models():
        if issubclass(model, BusinessModelObject):
            for bmo in model.objects.all():
                logger.info('Saving all objects for %s' % bmo, extra={'corrId': 'ACCOUNTIFIE.GL'})
                try:
                    bmo.save()
                except Exception as e:
                    msg = 'Recalc Fail. %s. %s' % (bmo.__dict__, e)
                    logger.exception(msg, extra={'corrId': 'ACCOUNTIFIE.GL'})

    # delete any orphan tranlines
    for tl in TranLine.objects.all():
        try:
            bmo_obj = tl.content_type.get_object_for_this_type(id=tl.object_id)
            if bmo_obj is None:
                tl.delete()
        except ObjectDoesNotExist as e:
            tl.delete()


class BusinessModelObject(object):
    """Something else which has an effect on the GL.

    BMOs will typically create General Ledger entries when
    saved.  They should efficiently update/change their
    GL entries and not create database churn unless there
    is a change.

    Make other models inherit from this as a mixin.

    Regrettably one then needs to override save() and
    delete() for the other models, until I find a cunning
    way to make it happen.

    """

    def delete_from_gl(self):
        "Find and remove all GL entries which depend on self"
        if get_gl_strategy() == 'remote':
            bmo_id = '%s.%s' %(self.short_code, self.id)
            QMSF.getInstance() \
                .get(strategy='remote') \
                .delete_bmo_transactions(self.company.id, bmo_id)
        elif GL_STRATEGY == 'postgres':
            TranLine.objects \
                    .filter(bmo_id='%s.%s' % (self.short_code, self.id)) \
                    .delete()

    def create_gl_transactions(self, trans):
        gl_strategy = get_gl_strategy()

        "Make any GL transactions this needs"
        for td in trans:
            try:
                comment = td['comment'][:99]
            except:
                comment = ''

            # old style
            if GL_STRATEGY == 'postgres':
                lines = [{'company_id': _model_to_id(td['company']),
                          'date': date,
                          'date_end': date,
                          'comment': comment,
                          'account_id': _model_to_id(account),
                          'amount': Decimal("{0:.2f}".format(amount)),
                          'counterparty_id': _model_to_id(counterparty),
                          'tags': ','.join(tags or []),
                          'bmo_id': td['bmo_id'],
                          'source_object': self,
                          'closing_entry': td.get('closing_entry', False)
                          } for account, date, amount, counterparty, tags in td['lines']]
                if sum(l['amount'] for l in lines) == DZERO:
                    db_tl_set = TranLine.objects.filter(bmo_id=td['bmo_id'])
                    new_tl_set = [TranLine(**l) for l in lines]

                    if db_tl_set.count() == 0:
                        save_tranlines(None, new_tl_set)
                    else:
                        db_lines = [tl._to_dict() for tl in db_tl_set]
                        new_lines = [tl._to_dict() for tl in new_tl_set]
                        chgs = DeepDiff(db_lines, new_lines)
                        
                        if len(chgs) > 0:
                            # if any changes then just delete/set to historical
                            save_tranlines(db_tl_set, new_tl_set)
                else:
                    logger.error('Imbalanced GL entries for %s' % td['bmo_id'])
            elif GL_STRATEGY == 'remote':
                d2 = td.copy()
                if 'date_end' not in d2:
                    d2['date_end'] = d2['date']

                try:
                    tags = ','.join(d2.pop('tags'))
                except:
                    tags = ''

                d2['company'] = d2['company'].id if isinstance (d2['company'], Company) else d2['company']

                lines = d2.pop('lines')
                trans_id = d2.pop('trans_id')
                bmo_id = d2.pop('bmo_id')

                lines = [{'account': account.id if isinstance (account, Account) else account,
                          'amount': "{0:.2f}".format(amount),
                          'counterparty': counterparty.id if isinstance (counterparty, Counterparty) else counterparty,
                          'tags': tags
                          } for account, amount, counterparty, tags in lines]

                QMSF.getInstance() \
                    .get(strategy='remote') \
                    .create_gl_transactions(d2, lines, trans_id, bmo_id)

    def update_gl(self):
        "Fix up GL after any kind of change"
        trans = self.get_gl_transactions()
        self.delete_from_gl()
        self.create_gl_transactions(trans)

    def get_gl_transactions(self):
        """Override this.  Create a list of transactions.

        Create a list of transactions in dictionary form
        which should exist.  It will create them all in the
        GL on saving if they do not already exist.
        """
        return []

    def get_company(self):
        return self.company.id


@transaction.atomic
def save_tranlines(old_tl_set, new_tl_set):
    if old_tl_set:
        old_tl_set.delete()

    for l in new_tl_set:
        try:
            l.save()
        except Exception as e:
            msg = 'Error saving tranlines. %s. %s' % (e, l.__dict__)
            logger.exception(msg, extra={'corrId': 'ACCOUNTIFIE.GL'})

def on_bmo_save(sender, **kwargs):
    """Alternative method - maintain GL through a signal

    Master-detail BMOs can't use a simple save hook, and
    instead need to set a signal to call this'
    """
    sender.update_gl()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-27 01:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cal', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='month',
            name='id',
            field=models.CharField(help_text='e.g. "2009M04"', max_length=7, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='quarter',
            name='id',
            field=models.CharField(help_text='e.g. "2009Q1"', max_length=6, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='year',
            name='id',
            field=models.IntegerField(help_text='4 digit year e.g. 2009', primary_key=True, serialize=False),
        ),
    ]



"""
Adapted with permission from ReportLab's DocEngine framework
"""


from django.conf.urls import url
from .views import *

urlpatterns = [
    url(r'^logs_data/$', db_logs_modal, name='dashboard_logs'),
    url(r'^logs/$', logs, name='logs'),
    url(r'^load_test/$', load_test, name='load_test'),
    url(r'^show_request/$', show_request, name='show_request'),
    url(r'^load_test_ui/$', load_test_ui, name='load_test_ui'),
    url(r'^stats/(?P<app_name>[0-9a-zA-Z_\.]+)/(?P<stat_name>\w+)/$', stat, name='dashboard_stat'),
    url(r'^(\w+)/$', app_index, name='dashboard_app'),
    url(r'^$', index, name="dashboard_index"),
]    



import json
from ast import literal_eval
from os.path import join as _join

from django.conf import settings
from django.template import Library

register = Library()


@register.filter
def company(args):
    try:
        _args = json.loads(args)
    except:
        _args = literal_eval(args)


    try:
        return _args[1]
    except:
        return 'unknown'

@register.filter
def file_path(args):
    cpny = company(args)
    return _join(settings.DATA_URL, settings.PDFOUT_PATH,
                   'report_pack_%s.pdf' % cpny)



from django.contrib import admin

from djcelery.models import TaskMeta


class TaskMetaAdmin(admin.ModelAdmin):
    list_display = ('task_id', 'date_done', 'result', 'status')
    readonly_fields = ('result',)    
admin.site.register(TaskMeta, TaskMetaAdmin)




from . import gl_helpers

import datetime
import json



from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from decimal import Decimal, getcontext, ROUND_HALF_UP, ROUND_HALF_DOWN
from bisect import insort
import locale
import operator
import re

import csv, os
from pprint import pprint

from django.db import models
from django.conf import settings

from . import datefuncs

locale.setlocale(locale.LC_ALL, '')



import logging
logger = logging.getLogger('default')


#use this frequently in rounding so give it a name
DZERO = Decimal("0.00")
HUNDREDTH = Decimal("0.01")

"""
def periods(period_tags):
    periods = {}
    for tag in period_tags:
        period_tag = period_tags[tag]
        periods[tag] = {'start': start_of_period(period_tag), 'end': end_of_period(period_tag)}
    return periods
"""

def day_before(d):
    return d - datetime.timedelta(days=1)

from pandas.tseries.offsets import BDay
def prev_busday(d):
    return (d - BDay(1)).date()


def entry(x, link='', ccy_fmt=''):
    return {'text': fmt(x, values_fmt=ccy_fmt), 'link': link}

def acct_history_link(id):
    from_dt = datefuncs.start_of_prev_year(datetime.datetime.now().year).isoformat()
    return ('/reporting/history/account/%s/?from=%s' % (id, from_dt))

def path_history_link(path):
    from_dt = datefuncs.start_of_prev_year(datetime.datetime.now().year).isoformat()
    return ('/reporting/history/path/%s/?from=%s' % (path.replace('.', '_'), from_dt))

def path_balances_link(path, date='today'):
    return ('/gl/path/%s/balances/?date=date' % path.replace('.','_'))

def fmt(stuff, values_fmt=None):
    "How do we want money formatted?"
    if stuff is None:
        return '----'
    elif isinstance(stuff, str):
        return stuff
    else:
        #accounting format, hacky, must be a recipe for this.
        #just seeing if brackets line up.
        if abs(stuff) < 0.1:
            txt = '-'
        elif stuff < 0:
            txt = '(' + locale.format("%0.0f", abs(stuff), grouping=True) + ')'
        else:
            txt = locale.format("%0.0f", stuff, grouping=True) + ''
        if values_fmt:
            return values_fmt + txt
        else:
            return txt

def unfmt(x):
    x = x.replace('$', '')
    if x == '-':
        return 0.0
    else:
        return float(x.replace(',','').replace(')','').replace('(','-'))





def get_dates(dt):
    if datefuncs.is_period_id(dt):
        start = datefuncs.start_of_period(dt)
        end = datefuncs.end_of_period(dt)
    elif type(dt) in [str, str] and dt[-4:]=='_YTD':
        end = parse(dt[:-4])
        start = datetime.date(end.year, 1, 1)
    elif type(dt) in [str, str] and dt[0] == 'D':
        d = parse(dt[1:]).date()
        start = prev_busday(d)+datetime.timedelta(days=1)
        end = d
    elif dt=='today':
        start = settings.DATE_EARLY
        end = datetime.datetime.now().date()
    else:
        start = settings.DATE_EARLY
        if type(dt)==str:
            end = parse(dt).date()
        else:
            end = dt
    return start, end

def get_dates_dict(dt):
    if type(dt)==dict:
        if 'start' in dt and 'end' in dt:
            #already in format
            return dt

    start, end = get_dates(dt)
    return { 'start': start, 'end': end }




def to_dict(dataset):
    "Make a dict with column 1 as key and cols 2,3,4 etc as value."
    d = {}
    for row in dataset:
        first, rest = row[0], row[1:]
        d[first] = rest
    return d

def denoneify(num):
    if num is None:
        return DZERO
    else:
        return num

def safe_sum(seq):
    "Safe with None"
    try:
        return sum(seq)
    except TypeError:
        seq2 = []
        for elem in seq:
            if elem is not None:
                seq2.append(elem)

        return sum(seq2)







def get_columns(request):
    if 'columns' in request.GET:
        return request.GET.get('columns').split('.')
    else:
        return None



def files_for_dir(datadir):
    return [os.path.join(datadir, name) for name in os.listdir(datadir) if os.path.isfile(os.path.join(datadir, name))]

def csv_to_modelattr(open_file, name_cleaner=None, company=gl_helpers.get_default_company()):
    '''takes the fields and values in the CSV and transforms them into a list of dicts where the keys
     will match model attributes. for example Full Description becomes full_description'''
    if name_cleaner == None:
        name_cleaner = lambda name: name
    f_csv = csv.DictReader(open_file)
    csv_to_modelattr = dict([(name, name_cleaner(name)) for name in f_csv.fieldnames])
    csv_to_modelattr['company_id'] = company

    return [dict([(csv_to_modelattr[name], value) for name, value in list(row.items()) if name in csv_to_modelattr]) for row in f_csv]

def get_foreignkeys(model):
    return dict(((f.name, f.rel.to) for f in model._meta.fields if f.__class__ == models.ForeignKey))

def get_fk_attr(model):
    return [f.name for f in model._meta.fields if f.__class__ == models.ForeignKey]

def get_pk_name(model):
    return model._meta.pk.name

def instance_nonrel_data(row, model, name_cleaner=None, value_cleaner=None):
    model_flds =  model._meta.get_all_field_names()
    instance_data_no_fk = dict((name_cleaner(name), value_cleaner(name, value)) for name, value in list(row.items()) if name_cleaner(name)
                        and name_cleaner(name) not in get_fk_attr(model) and name_cleaner(name) in model_flds)
    return model(**instance_data_no_fk)

def set_foreignkeys(instance, row, model, name_cleaner=None, value_cleaner=None):
    if get_foreignkeys(model):
        instance_fk = dict((name_cleaner(name), value_cleaner(name, value)) for name, value in list(row.items()) if name_cleaner(name)
                        and name_cleaner(name) in get_fk_attr(model))
        for fk in list(get_foreignkeys(model).items()):
            if fk[0] in instance_fk:
                try:
                    related = fk[1].objects.get(pk=instance_fk[fk[0]])
                    setattr(instance, fk[0], related)
                except:
                    logger.error("No ForeignKey %s %s.  %s" % (fk[0], str(fk[1]), instance_fk))
    return instance

def dirty_key(row, model=None, unique=None, name_cleaner=None, value_cleaner=None):
    dirty = [name_cleaner(k) for k in list(row.keys())
                if name_cleaner(k) not in [f.name for f in [field for field in model._meta.fields
                        if field not in get_fk_attr(model)]]
                                if name_cleaner(k)]

    return dirty

def create_instance(row, model, name_cleaner=None, value_cleaner=None, unique=None, exclude=[], company=gl_helpers.get_default_company()):
    row['company'] = company
    non_rel_instance = instance_nonrel_data(row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner)
    if non_rel_instance.id in exclude:
        return None

    full_instance = set_foreignkeys(non_rel_instance, row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner)

    return unique(full_instance)


def random_color(dark=True):
    import random
    r = lambda: random.randrange(10,130 if dark else 255, 10)
    return '#%02X%02X%02X' % (r(),r(),r())



import inspect
import importlib

from django.core.management.base import BaseCommand, CommandError
from django.apps import apps as django_apps
from django.db import transaction

from accountifie.gl.bmo import BusinessModelObject
from accountifie.gl.models import TranLine
from accountifie.common.api import api_func

import logging
logger = logging.getLogger('default')


class Command(BaseCommand):

    help = 'Rebuild ALL general ledger entries'
    
    def handle(self, *args, **options):
        verbosity = options["verbosity"]
        if verbosity > 0:
            print("Dropping general ledger and rebuilding all entries")

        """
        klasses = []
        kl_paths = api_func('environment', 'variable_list', 'BMO_MODULES')

        # find all the BMO classes
        for path in kl_paths:
            for name, kl in inspect.getmembers(importlib.import_module(path), inspect.isclass):
                if BusinessModelObject in kl.__bases__:
                    klasses.append(kl)

        
        with transaction.atomic():
            #quicker to do the two delets ourselves than let the engine work out all the cascading stuff
            TranLine.objects.all().delete()
        """

        for model in django_apps.get_models():
            if issubclass(model, BusinessModelObject):
                if verbosity > 0:
                    print("rebuilding GL entries from %s.%s" % (model._meta.app_label, model.__name__))

                with transaction.atomic():
                    for bmo in model.objects.all():
                        try:
                            bmo.save()
                        except:
                            logger.error('rebuild_postgres_gl: failed on %s' % bmo)



"""
Adapted with permission from ReportLab's DocEngine framework
"""



# as this is in a models.py, it will get loaded at django server start
# it intends to fail 
"""
import os, sys
from django.conf import settings

tcp = settings.TEMPLATE_CONTEXT_PROCESSORS


if 'accountifie.common.views.base_templates' not in tcp:
    sys.stderr.write('''!!!
    You have upgrade DocEngine
    Please add:

TEMPLATE_CONTEXT_PROCESSORS = (
    "django.contrib.auth.context_processors.auth",
    "django.core.context_processors.debug",
    "django.core.context_processors.i18n",
    "django.core.context_processors.media",
    "django.core.context_processors.static",
    "django.core.context_processors.tz",
    "django.contrib.messages.context_processors.messages",
    "accountifie.common.views.base_templates",
    )

    to your settings.py
''')
    os.kill(os.getpid(),9)
    sys.exit(1)
"""


# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-12-21 02:38
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0004_tranline_date'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tranline',
            name='counterparty',
            field=models.ForeignKey(blank=True, max_length=50, null=True, on_delete=django.db.models.deletion.CASCADE, to='gl.Counterparty'),
        ),
    ]



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Day',
            fields=[
                ('id', models.DateField(serialize=False, primary_key=True)),
            ],
        ),
        migrations.CreateModel(
            name='Month',
            fields=[
                ('id', models.CharField(help_text=b'e.g. "2009M04"', max_length=7, serialize=False, primary_key=True)),
            ],
        ),
        migrations.CreateModel(
            name='Quarter',
            fields=[
                ('id', models.CharField(help_text=b'e.g. "2009Q1"', max_length=6, serialize=False, primary_key=True)),
            ],
        ),
        migrations.CreateModel(
            name='Week',
            fields=[
                ('id', models.CharField(max_length=7, serialize=False, primary_key=True)),
                ('first_day', models.DateField(null=True)),
                ('last_day', models.DateField(null=True)),
                ('end_month', models.ForeignKey(related_name='end_week_set', to='cal.Month', on_delete=models.CASCADE)),
                ('start_month', models.ForeignKey(related_name='start_week_set', to='cal.Month', on_delete=models.CASCADE)),
            ],
        ),
        migrations.CreateModel(
            name='Year',
            fields=[
                ('id', models.IntegerField(help_text=b'4 digit year e.g. 2009', serialize=False, primary_key=True)),
            ],
        ),
        migrations.AddField(
            model_name='week',
            name='year',
            field=models.ForeignKey(to='cal.Year', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='quarter',
            name='year',
            field=models.ForeignKey(to='cal.Year', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='month',
            name='quarter',
            field=models.ForeignKey(to='cal.Quarter', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='month',
            name='year',
            field=models.ForeignKey(to='cal.Year', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='day',
            name='month',
            field=models.ForeignKey(to='cal.Month', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='day',
            name='quarter',
            field=models.ForeignKey(to='cal.Quarter', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='day',
            name='year',
            field=models.ForeignKey(to='cal.Year', on_delete=models.CASCADE),
        ),
    ]



# Generated by Django 2.0.5 on 2018-06-16 20:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0009_auto_20180607_1718'),
    ]

    operations = [
        migrations.AddField(
            model_name='tranline',
            name='closing_entry',
            field=models.BooleanField(default=False),
        ),
    ]



"""
This file demonstrates two different styles of tests (one doctest and one
unittest). These will both pass when you run "manage.py test".

Replace these with more appropriate tests for your application.
"""

from django.test import TestCase
from accountifie.cal import *
from accountifie.cal.models import Day, Month, Year, Week, Quarter
class SimpleTest(TestCase):
    def test_basic_addition(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        self.assertEqual(1 + 1, 2)


class CalendarTests(TestCase):
    def test_populate_calendar(self):
        populate()
        self.assertTrue(Day.objects.count() in [365, 366])
        self.assertTrue(Week.objects.count() in [52, 53, 54])
        self.assertEqual(Month.objects.count(), 12)
        self.assertEqual(Quarter.objects.count(), 4)
        self.assertEqual(Year.objects.count(), 1)

    def test_regex_periods(self):
        self.assertTrue(PAT_YEAR.match('2014'))
        self.assertFalse(PAT_YEAR.match('99'))
        self.assertFalse(PAT_YEAR.match('12345'))

        self.assertTrue(PAT_QUARTER.match('2014Q1'))
        self.assertTrue(PAT_QUARTER.match('2014Q4'))
        self.assertFalse(PAT_QUARTER.match('2014Q5'))

        self.assertTrue(PAT_HALF.match('2014H1'))
        self.assertTrue(PAT_HALF.match('2014H2'))
        self.assertFalse(PAT_HALF.match('2014H3'))

        self.assertTrue(PAT_MONTH.match('2014M01'))
        self.assertTrue(PAT_MONTH.match('2014M12'))

        self.assertFalse(PAT_MONTH.match('2014M7'))
        self.assertFalse(PAT_MONTH.match('2014M00'))
        self.assertFalse(PAT_MONTH.match('2014M13'))


        self.assertTrue(PAT_WEEK.match('2014W01'))
        self.assertFalse(PAT_WEEK.match('2014W00'))
        self.assertFalse(PAT_WEEK.match('2014W54'))



    def test_period_ids(self):
        "Useful as query params, check if valid date range"
        for thing in ['2014Q1', '2013M07', '2012Q4', '2014W36']:
            self.assertTrue(is_period_id(thing))

        for thing in [None, '', 201, 12345, '99', '2013M00', '2012Q5', '2014W56', '2014H0']:
            self.assertFalse(is_period_id(thing), "is_period_id('%s') succeeded unexpectedly" % thing)


        self.assertTrue(is_iso_date("2014-08-31"))
        self.assertFalse(is_iso_date("31/8/2014"))

    def test_start_of_period(self):
        
        for thing, expected in [
            ['2014M01', date(2014,1,1)],
            ['2014M03', date(2014,3,1)],
            ['2014M04', date(2014,4,1)],
            ['2014M12', date(2014,12,1)],

            ['2014Q1', date(2014,1,1)],
            ['2014Q2', date(2014,4,1)],
            ['2014Q3', date(2014,7,1)],
            ['2014Q4', date(2014,10,1)],

            ['2014H1', date(2014,1,1)],
            ['2014H2', date(2014,6,1)],

            ['2014', date(2014,1,1)],
            ]:
            self.assertEqual(start_of_period(thing), expected)


    def test_end_of_period(self):
        
        for thing, expected in [
            ['2014M01', date(2014,1,31)],
            ['2014M03', date(2014,3,31)],
            ['2014M04', date(2014,4,30)],
            ['2014M12', date(2014,12,31)],

            ['2014Q1', date(2014,3,31)],
            ['2014Q2', date(2014,6,30)],
            ['2014Q3', date(2014,9,30)],
            ['2014Q4', date(2014,12,31)],

            ['2014H1', date(2014,6,30)],
            ['2014H2', date(2014,12,31)],

            ['2014', date(2014,12,31)],
            ]:
            self.assertEqual(end_of_period(thing), expected)


__test__ = {"doctest": """
Another way to test that 1 + 1 is equal to 2.

>>> 1 + 1 == 2
True

Calendar tests:

>>> from docengine.cal import previous_period, next_period
>>> previous_period("2014W22")
"2014W21"
>>> previous_period("2014W01")
"2013W52"
>>> previous_period("2014M01")
"2013M12"
>>> previous_period("2014Q1")
"2013Q4"
>>> previous_period("2014H1")
"2013H2"

>>> next_period("2014W22")
"2014W23"
>>> next_period("2014W52")
"2015W01"
>>> next_period("2014Q4")
"2015Q1"
>>> next_period("2014H2")
"2015H1"
>>> next_period("2014M12")
"2015M01"




"""}




import json
import logging
import importlib


logger = logging.getLogger('default')


def bstrap_table(table_name):
    try:
        tbl_module = importlib.import_module('tables.bstrap_tables')
        return getattr(tbl_module, table_name)
    except:
        pass

    try:
        tbl_module = importlib.import_module('accountifie.reporting.bstrap_tables.common')
        return getattr(tbl_module, table_name)
    except:
        logger.error("couldn't find bootstrap table def %s" % table_name)
        return None




# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-06-17 03:34


from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Address',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('street_number', models.CharField(blank=True, max_length=20)),
                ('route', models.CharField(blank=True, max_length=100)),
                ('postal_code', models.CharField(blank=True, max_length=10)),
                ('raw', models.CharField(max_length=200)),
                ('formatted', models.CharField(blank=True, max_length=200)),
                ('latitude', models.FloatField(blank=True, null=True)),
                ('longitude', models.FloatField(blank=True, null=True)),
            ],
            options={
                'verbose_name_plural': 'Addresses',
                'db_table': 'address_address',
                'ordering': ('locality', 'route', 'street_number'),
            },
        ),
        migrations.CreateModel(
            name='Country',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=70, unique=True)),
                ('code', models.CharField(blank=True, max_length=5)),
            ],
            options={
                'verbose_name_plural': 'Countries',
                'db_table': 'address_country',
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='Locality',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=165)),
            ],
            options={
                'verbose_name_plural': 'Localities',
                'db_table': 'address_locality',
                'ordering': ('state', 'name'),
            },
        ),
        migrations.CreateModel(
            name='State',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=165)),
                ('code', models.CharField(blank=True, max_length=3)),
                ('country', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='states', to='address.Country')),
            ],
            options={
                'db_table': 'address_state',
                'ordering': ('country', 'name'),
            },
        ),
        migrations.AddField(
            model_name='locality',
            name='state',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='localities', to='address.State'),
        ),
        migrations.AddField(
            model_name='address',
            name='locality',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='addresses', to='address.Locality'),
        ),
        migrations.AlterUniqueTogether(
            name='state',
            unique_together=set([('name', 'country')]),
        ),
        migrations.AlterUniqueTogether(
            name='locality',
            unique_together=set([('name', 'state')]),
        ),
    ]



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0002_address'),
    ]

    operations = [
        migrations.AddField(
            model_name='address',
            name='label',
            field=models.CharField(default='Default', max_length=20),
            preserve_default=False,
        ),
    ]



"""
copied and amended from https://github.com/furious-luke/django-address
"""

from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from .models import *

class UnidentifiedListFilter(SimpleListFilter):
    title = 'unidentified'
    parameter_name = 'unidentified'

    def lookups(self, request, model_admin):
        return (('unidentified', 'unidentified'),)

    def queryset(self, request, queryset):
        if self.value() == 'unidentified':
            return queryset.filter(locality=None)

@admin.register(Country)
class CountryAdmin(admin.ModelAdmin):
    search_fields = ('name', 'code')

@admin.register(State)
class StateAdmin(admin.ModelAdmin):
    search_fields = ('name', 'code')

@admin.register(Locality)
class LocalityAdmin(admin.ModelAdmin):
    search_fields = ('name',)

@admin.register(Address)
class AddressAdmin(admin.ModelAdmin):
    search_fields = ('name',)
    list_filter = (UnidentifiedListFilter,)



from django.conf.urls import include, url
from django.contrib import admin

admin.autodiscover()

urlpatterns = [

    #url(r'^admin/', include(admin.site.urls)),

    url(r'', include('accountifie.toolkit.urls')),
    url(r'', include('accountifie.common.urls')),
    url(r'', include('accountifie.celery.urls')),
    url(r'', include('accountifie.gl.urls')),

]



"""
Adapted with permission from ReportLab's DocEngine framework
"""


import os
import sys
import time
import json 
import logging

from datetime import datetime
from optparse import make_option
try:
    from splinter import Browser
except:
    sys.stdout.write('You need to install splinter for this to run properly!!!\n')
    sys.exit(1)

from time import clock
from urllib.parse import urlparse

from django.conf import settings
from django.core.mail import send_mail
from django.core.management.base import BaseCommand
from django.template.defaultfilters import slugify

logger = logging.getLogger('default')

LOGS_DIR = os.path.realpath(os.path.join(settings.PROJECT_DIR, '..', 'logs'))

class Command(BaseCommand):
    """
    Logs the access time for a given URL as JSON under <repository>/logs/
    (defaults to http://www.reportlab.com/)
    filename will be made of the `slugify(last-10-chars-of-url)`.json
    """
    help = __doc__
    option_list = BaseCommand.option_list + (
        make_option("-u", "--url", 
            dest="url",
            help="The URL (with port) you want to test against",
            default="http://www.reportlab.com/"
            ),
        make_option("-b", "--browser", 
            dest="browser",
            help="Choose the browser: chrome or phantomjs", 
            default="phantomjs"
            ),
        make_option("-s", "--search", 
            dest="search",
            help="Search string to find in the content of the resulting page", 
            default=""
            ),
        make_option("-t", "--threshold", 
            dest="threshold",
            help="Last N log entries with go over the limit", 
            default=5
            ),
        make_option("-l", "--limit", 
            dest="limit",
            help="Time limit in seconds above which an error will be reported",
            default=5
            )
    )

    def handle(self, *args, **options):
        verbose = int( options['verbosity'])
        browser = options['browser']
        url = options.get('url', "http://www.reportlab.com/")
        search = options['search']
        limit = int(options['limit']*1000)
        threshold = int(options['threshold'])


        # naming sjon log upon the provided url
        if not os.path.exists(LOGS_DIR):
            os.makedirs(LOGS_DIR)
        filename = os.path.join(LOGS_DIR, "%s.json" % slugify(url[-10:]).replace('-','_'))
        screenshotname = os.path.splitext(filename)[0]
            


        if '?' not in url and ('/media/' or '/static/' not in url) and not url.endswith('/'):
            url+='/'
        with Browser(browser) as b:
            for i in range(3):
                started = clock()
                b.visit(url)
                finished = time.clock()
                print('page visit in %0.3f seconds' % (finished - started))
            if not b.status_code.is_success:
                logging.error('[time_page] Request to %s returned an Error' % search)
                if verbose>1:
                    self.stdout.write('[time_page] Request to %s returned an Error' % search)
            if search and not search in b.html:
                logging.error('%s could not be found in the resulting page' % search)
                if verbose>1:
                    self.stdout.write('%s could not be found in the resulting page' % search)
                return 1
            
            how_long = int(1000 * (finished - started))
            if verbose > 1:
                print(('%0.2fms\n' % how_long))
                # sys.exit(1)
            if os.path.isfile(filename):
                with open(filename, 'r') as infile:
                    try:
                        log = json.load(infile)
                    except:
                        log = []
            else:
                log = []
            log.append({'when': datetime.now().strftime("%a, %d %b %Y %H:%M:%S +0000"),
                    'how_long': how_long}
                    )
            with open(filename, 'w') as outfile:
                json.dump(log, outfile)

            last_ones = log[-threshold:]
            exceeds_limit = True 
            for i in last_ones:
                if i['how_long'] < threshold * 1000:
                    exceeds_limit = False
                    break
            if exceeds_limit:
                logger.error('Excessive loading time (>%sms) accessing url %s' % (limit, url))




"""
LoggingMiddleware used by Hilton and others to generate
our traditional log files and capture most details of the
request.

"""

try:
    from threading import local
except ImportError:
    from django.utils._threading_local import local

_thread_locals = local()

def getCurrentUser():
    "Call this from anywhere to get the user, assuming middleware below is installed"
    return getattr(_thread_locals, 'user', None)

def getCurrentRequest():
    "Call this from anywhere to get the request, assuming middleware below is installed"
    return getattr(_thread_locals, 'request', None)

class UserFindingMiddleware(object):
    """Middleware that gets various objects from the
    request object and saves them in thread local storage."""
    def process_request(self, request):
        _thread_locals.user = getattr(request, 'user', None)
        _thread_locals.request = request

from django.conf import settings
class ReplaceText(object):
    '''
    settings.TEXT_REPLACEMENTS = (
        'text0', 'replacement0',
        'text1', 'replacement1',
        ........
        'textn', 'replacementn',
        )
    replaces texti with replacementi in response.content

    *NB* this is not clever about substrings so do longer replacements first.
    '''
    def process_response(self, request, response):
        content = response.content
        TR = getattr(settings,'TEXT_REPLACEMENTS',())
        for i in range(0,len(TR),2):
            content = content.replace(TR[i],TR[i+1])
        response.content = content
        return response



from django.conf import settings
from django.conf.urls import url, static

from . import views
from . import jobs


urlpatterns = [
    url(r'^choose_company/(?P<company_id>.*)/$', views.choose_company, name='choose_company'),    
    url(r'^cleanlogs/$', jobs.cleanlogs, name='cleanlogs'),
    url(r'^primecache/$', jobs.primecache, name='primecache'),
    url(r'^dump_fixtures/$', views.dump_fixtures),
]    




"""
Strategy that implements basic GL actions for test suite.
See query_manager_strategy.py for interface docs.
Use query_manager_strategy_factory.py to get an instance of this class.
"""

from .query_manager_strategy import QueryManagerStrategy


class QueryManagerTestStrategy(QueryManagerStrategy):

  def create_gl_transactions(self, d2, lines, trans_id, bmo_id):
    pass


  def delete_bmo_transactions(self, company_id, bmo_id):
    pass



# Generated by Django 2.0.5 on 2018-06-07 21:18

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0008_auto_20180511_1529'),
    ]

    operations = [
        migrations.AlterField(
            model_name='company',
            name='label',
            field=models.CharField(default='TBD', max_length=20),
        ),
        migrations.AlterField(
            model_name='counterparty',
            name='id',
            field=models.CharField(max_length=30, primary_key=True, serialize=False, validators=[django.core.validators.RegexValidator('^[0-9a-zA-Z]*$', 'Only alphanumeric characters are allowed. In particular.. no spaces.')]),
        ),
        migrations.AlterField(
            model_name='counterparty',
            name='label',
            field=models.CharField(default='TBD', max_length=20),
        ),
    ]



from django.conf.urls import *

from . import views
from . import jobs


urlpatterns = [

    # admin-type view
    url(r'gl/accounts_list/$', views.accounts_list),
    url(r'gl/counterparty_list/$', views.counterparty_list),

    # utilities
    url(r'gl/get_transactions/', views.download_transactions),
    url(r'gl/get_tranlines/', views.download_tranlines),

    url(r'celery/recalculate/', jobs.recalculate),
]



"""
Adapted with permission from ReportLab's DocEngine framework
"""

import re, os, sys
from collections import OrderedDict
import json

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.admin.views.decorators import staff_member_required
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.db import connection
from django.shortcuts import render
from django.template import RequestContext
from django.http import Http404, HttpResponse, HttpResponseBadRequest, HttpResponseForbidden, HttpResponseNotAllowed, HttpResponseServerError
from django.views.decorators.csrf import csrf_protect, csrf_exempt
from accountifie.common.models import Log
from accountifie.common.dbutils import all_relations, all_concrete_models

from . import utils


LOG_ENTRIES_PER_PAGE = 25

def localsets():
    res = {}
    project_name = os.path.basename(settings.PROJECT_DIR)
    try:
        ls = __import__('%s.localsettings' % project_name)
        for k, v in list(ls.localsettings.__dict__.items()):
            if k == k.upper() and re.match("[^__]", k):
                res[k]=v
    except:
        pass
    return res

def settings_list():
    res = []
    omit_list = ('SECRET_KEY', 'DB_PASSWORD',)
    for ke, va in [ (k, v) for k, v in list(settings._wrapped.__dict__.items()) if not k.startswith('_') and k not in omit_list]:
        if ke=='DATABASES':
            for name, props in list(settings.DATABASES.items()):
                res.append(('DATABASE "%s"' % name, props['NAME']),)
        else:
            if va:
                res.append((ke, va),)
    return res

@staff_member_required
def index(request):
    pip = os.path.join(sys.exec_prefix, 'bin', 'pip')
    if not os.path.isfile(pip):
        pip = 'pip'
    SHELL_COMMANDS = (
        ('Hostname','hostname'),
        ('hg version', 'hg id'),
        ('git version', "git log --pretty=format:'%h' -n 1"),
        ('hg branch', 'hg branch'),
        ('git branch', 'git rev-parse --abbrev-ref HEAD'),
        ('MySQL version', 'mysql --version'),
        ('Local Packages', '%s freeze -l' % pip)
    )
    SD = OrderedDict()
    for k,v in sorted(settings_list(), key=lambda x: x[0]):
        SD[k] = v
    context = RequestContext(request, {
        'args': sys.argv,
        'exe': sys.executable,
        'settings': SD,
        })

    context['versions'] = OrderedDict()
    # get versions
    curr_dir = os.path.realpath(os.path.dirname(__file__))
    for name, shell_command in SHELL_COMMANDS:
        try:
            result = utils.run_shell_command(shell_command, curr_dir)
            if result:
                if isinstance(result, list):
                    result = '<br>'.split(result)
                context['versions'][name] = result
        except:
            pass
    # machine status
    context['machine'] = OrderedDict()
    if sys.platform == 'darwin':
        context['machine']['Uptime'] = 'not done yet on MacOS'
        context['machine']['Disk Space'] = 'not done yet on MacOS'
    elif sys.platform == 'win32':
        context['machine']['Uptime'] = 'not done yet on Windows'
        context['machine']['Disk Space'] = 'not done yet on Windows'
    else:
        context['machine']['Uptime'] = utils.server_uptime()
        context['machine']['Disk Space'] = utils.disk_usage('/')._asdict()
    if os.path.exists(settings.MEDIA_ROOT):
        context['machine']['Media Folder'] = utils.sizeof_fmt(utils.folder_size(settings.MEDIA_ROOT))

    context['stats'] = utils.get_available_stats()

    gan = lambda app: app
    context['apps'] = [(gan(app), ', '.join([model.__name__ for model in models])) for app, models in all_concrete_models()]
    context['relations'] = [[(model.__name__, ', '.join(['%s (%s) through %s' % (relation.__name__, relation.__module__, field.__class__.__name__)
                                                        for field, relation in relations]), gan(app))
                                                            for model, relations in rel_info]
                                                                for app, rel_info in all_relations()]
    #context['rel_graph'] =

    context['config_warnings'] = utils.get_configuration_warnings()

    return render(request, 'dashboard/index.html', context)

@staff_member_required
def logs(request):
    level = request.GET.get('level')
    return render(request, 'dashboard/logs.html', {'level': level})

@login_required
def db_logs_modal(request):
    level = request.GET.get('level')

    if level:
        log_list = Log.objects.filter(level=level)
    else:
        log_list = Log.objects.all()
    paginator = Paginator(log_list, LOG_ENTRIES_PER_PAGE) # Show 25 contacts per page

    page = request.GET.get('page')
    try:
        logs = paginator.page(page)
    except PageNotAnInteger:
        # If page is not an integer, deliver first page.
        logs = paginator.page(1)
    except EmptyPage:
        # If page is out of range (e.g. 9999), deliver last page of results.
        logs = paginator.page(paginator.num_pages)

    return render(request, 'dashboard/logs_excerpt.html', {"logs": logs, "level": level})


@login_required
def app_index(request, app_name):
    '''
    This should query the app for any media related views
    '''
    try:
        app_utils = __import__('project.%s.utils' % app_name)
        return render(request, 'dashboard/app.html', {'assets': app_utils.media()})
    except:
        pass
        # no app builtin mode of retrieving user media

    context = RequestContext(request, {
        'app_name': app_name,
        'assets': {
            'images': get_media('images', app_name),
            'fonts': get_media('fonts', app_name)
        }
    })
    return render(request, 'dashboard/app.html', context)

@login_required
def stat(request, app_name, stat_name):
    '''
    Runs the designated canned query and shows the results
    '''

    from django.utils.importlib import import_module
    try:
        stats = import_module('.stats', app_name)
    except ImportError as err:
        raise Http404

    found = False
    if hasattr(stats, 'SQL_QUERIES'):
        for (name, help_text, sql) in stats.SQL_QUERIES:
            if name == stat_name:
                found = True
                cur = connection.cursor()
                cur.execute(sql)
                data =cur.fetchall()

                colnames = [desc[0] for desc in cur.description]
    if not found:
        raise Http404




    context = RequestContext(request, {
        'app_name': app_name,
        'stat_name': stat_name,
        'help_text': help_text,
        'colnames': colnames,
        'data': data
        })
    return render(request, 'dashboard/stat.html', context)




@login_required
@csrf_exempt
def load_test(request):
    if request.method == 'POST':
        number = request.POST['number']
        if os.environ.get('FLAKY'):
            from random import random
            rnd = random()
            if rnd > 0.8:
                return HttpResponseBadRequest()
            elif rnd > 0.6:
                return HttpResponseForbidden()
            elif rnd > 0.4:
                return HttpResponseNotAllowed()
            elif rnd > 0.2:
                return HttpResponseServerError()
            elif rnd > 0.4 and rnd < 0.6:
                return HttpResponse(json.dumps(dict(number=int(number)*2)))
        return HttpResponse(json.dumps(dict(number=int(number)*2)))

@login_required
def load_test_ui(request):
    return render(request, 'dashboard/load_tester_ui.html')

@login_required
def show_request(request):
    """Lets you inspect details about the request.

    May be useful when investigating web server settings and their effects.
    If you want to temporarily hook it up to handle all patterns, add a URL
    to match everything and route to it.
    """
    params = dict(
            request=request,
            env='\n'.join(['%s=%r' % (k,v) for k,v in os.environ.items()]),
            )
    return render(request, 'dashboard/show_request.html', params)




from django.shortcuts import render
from django.contrib.auth.decorators import login_required

from .apiv1 import tasks_in_progress
from django.http import JsonResponse

@login_required
def tasks_list(request):
    return JsonResponse(tasks_in_progress(), safe=False)



"""
Interface for swappable strategies used for dealing with transactions in QueryManager

Based on the strategy pattern https://en.wikipedia.org/wiki/Strategy_patter
"""
class QueryManagerStrategy(object):
  def account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra, with_tags, excl_tags):
    """
    :param company_id:           the company id
                                 e.g. 'INC'
    :param account_ids:          a list of account ids
                                 e.g. [ '3001', '3010' ]
    :param dates:                dictionary of date ranges
                                 e.g. { '2015': { 'start': '2015-01-01', 'end': '2015-12-31' } }
    :param with_counterparties:  list containing transaction counterparties that should be used to calculate the balance
                                 e.g. [ 'jpmc' ]
    :param excl_interco:         boolean indicating intercompany (e.g. 'INC' transaction with 'LLC' contra) transactions should
                                 be excluded if INC and LLC both subs of same company.
                                 e.g. True
    :param excl_contra:          a list of contra account ids to be excluded
                                 e.g. [ '3001' ]
    :param with_tags             a list of tags for which to find GL entries
                                 e.g. ['yearend']
    :param excl_tags             a list of tags for which to exclude GL entries
                                 e.g. ['yearend']

    :return:                     a dictionary of account balances, indexed by date
                                 e.g. { '2015': { '3001': '1503.23', '3010': '1626.23' } }
    """
    raise Exception('Unimplemented: account_balances_for_dates')

  def transactions(self, company_id, account_ids, from_date, to_date, chunk_frequency, with_counterparties, excl_interco, excl_contra):
    """
    :param company_id:          the company id
                                e.g. 'INC'
    :param account_ids:         a list of account ids
                                e.g. [ '3001', '7020' ]
    :param from_date:           get transactions starting on this date
                                e.g. '2015-01-01'
    :param to_date:             get transactions up to and including this date
                                e.g. '2015-01-31'
    :param chunk_frequency:     when to break spanning transactions (e.g. depreciation)
                                e.g. 'end-of-month'
    :param with_counterparties: list for filtering transaction list by a specific counterparty
                                e.g. [ 'jpmc' ]
    :param excl_interco:        boolean indicating intercompany (e.g. 'INC' transaction with 'LLC' contra) transactions should
                                be excluded if INC and LLC both subs of same company.
    :param excl_contra:         a list of contra account ids to be excluded
                                e.g. [ '3001' ]
    :return:                    a list of transactions
                                e.g. [
                                   { date: Date(2015-01-05), id: '270076', comment: '8661368: Egnyte', account_id: '7020', contra_accts: '3000', counterparty: 'egnyte', amount: 90.00  },
                                   { ... }, ...
                                 ]
    """
    raise Exception('Unimplemented: transactions')

  def upsert_transaction(self, transaction):
    """
    :param transaction:     the transaction object to update/create.
                            e.g. [{
                              'id': 12345,
                              'company': 'INC',
                              'comment': 'TEST TEST TEST',
                              'date': '2015-05-15',
                              'date_end': '2015-05-16',
                              'object_id': 1234,
                              'lines': [
                                { 'account': '1002', 'amount': '-10.00', 'counterparty': '(courier)' },
                                { 'account': '1001', 'amount': '10.00',  'counterparty': '(courier)' }
                              ]
                            }]
    :return: None
    """
    pass

  def delete_transaction(self, company_id, transaction_id):
    """
    :param transaction_id:  a string identifying the transaction to delete
    :return: None
    """
    pass

  def create_gl_transactions(self, d2, lines, trans_id, bmo_id):
    """
    :param
    :return: None
    """
    pass

  def delete_bmo_transactions(self, company_id, bmo_id):
    """
    :param bmo_id:  a string identifying the BMO for which to delete transactions
    :return: None
    """
    pass

  def erase(self, company_id):
    """
    :param company_id:      the company id
                            e.g. 'INC'
    :return: None
    """
    pass

  def set_fast_inserts(self, company_id, value):
    """
    :param company_id:      the company id
                            e.g. 'INC'
    :param value: Boolean
    :return:
    """
    pass

  def take_snapshot(self, company_id):
    """
    :param company_id:      the company id
                            e.g. 'INC'
    :return:
    """
    pass



# Project project Django settings.
import os, sys, json, pandas


# stop those annoying warnings
pandas.options.mode.chained_assignment = None

PROJECT_NAME = 'accountifie'
PROJECT_DIR = os.path.realpath(os.path.dirname(__file__))
sys.path.append(PROJECT_DIR)

ENVIRON_DIR = os.path.realpath(os.path.join(PROJECT_DIR, '..'))

CLIENT_PROJECT = os.path.split(ENVIRON_DIR)[1]

DEBUG = False

DEFAULT_GL_STRATEGY = os.environ.get('DEFAULT_GL_STRATEGY', 'remote')


# end of over-rides for ansible lineinfile

TEMPLATE_DEBUG = False

ADMINS = (
    ('admin', 'test@accountifie.com'),
)

MANAGERS = ADMINS

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(PROJECT_DIR, 'docengine_test.db'),
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}

TIME_ZONE = 'America/New_York'
LANGUAGE_CODE = 'en-gb'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/home/media/media.lawrence.com/media/"
MEDIA_ROOT = os.path.join(ENVIRON_DIR, 'media')
DATA_ROOT = os.path.join(ENVIRON_DIR, 'data')

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://media.lawrence.com/media/", "http://example.com/media/"
MEDIA_URL = '/media/'
DATA_URL = '/data/'

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/home/media/media.lawrence.com/static/"
STATIC_ROOT = os.path.join(ENVIRON_DIR, 'htdocs', 'static')

# URL prefix for static files.
# Example: "http://media.lawrence.com/static/"
STATIC_URL = '/static/'

PDFOUT_PATH = 'pdfout'
PDFOUT = os.path.join(DATA_ROOT, PDFOUT_PATH)


# Additional locations of static files
STATICFILES_DIRS = (
    os.path.join(PROJECT_DIR, 'static'),
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
#    'django.contrib.staticfiles.finders.DefaultStorageFinder',
)

# Make this unique, and don't share it with anybody.
SECRET_KEY = '_=s8f!l_t=ys+nbm3q%08ew8zb(7bybf195*rl2dil87p197g$'

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.app_directories.Loader',
    'django.template.loaders.filesystem.Loader',
#     'django.template.loaders.eggs.Loader',
)


MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'accountifie.middleware.docengine.UserFindingMiddleware',
    'djangosecure.middleware.SecurityMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'accountifie.middleware.ssl.SSLRedirect',
    'accountifie.toolkit.error_handling.StandardExceptionMiddleware'
)

TEMPLATE_CONTEXT_PROCESSORS = (
    "django.contrib.auth.context_processors.auth",
    "django.core.context_processors.debug",
    "django.core.context_processors.i18n",
    "django.core.context_processors.media",
    "django.core.context_processors.static",
    "django.core.context_processors.tz",
    "django.contrib.messages.context_processors.messages",
    "accountifie.common.views.base_templates",
    )


ROOT_URLCONF = PROJECT_NAME + '.urls'
WSGI_APPLICATION = PROJECT_NAME + '.wsgi.application'

TEMPLATE_DIRS = (
    # Put strings here, like "/home/html/django_templates" or "C:/www/django/templates".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
    os.path.join(PROJECT_DIR, 'templates'),

)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.humanize',
    # Uncomment the next line to enable admin documentation:
    # 'django.contrib.admindocs',
    'djangosecure',
    #'cerberos',
    'accountifie.dashboard',

    'django_nose',
    'django_extensions',

    'betterforms',
    #'json_field',

    'accountifie.common',
    'accountifie.cal',
    'accountifie.gl',
    'accountifie.environment',

    'django_admin_bootstrapped',
    'django.contrib.admin',


)

LOGIN_REDIRECT_URL = '/'
LOGIN_URL = '/login/'
LOGOUT_URL = '/logout/'


CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.dummy.DummyCache',
    }
}


# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s'
            },
        'simple': {
            'format': '%(levelname)s %(module)s %(message)s',
        },
        },
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
            }
        },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
            },
        'console':{
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
            },
        'database': {
            'level': 'INFO', # i.e., allows for logging messages of level INFO or higher
            'class': 'accountifie.common.log.DbLogHandler'
            }
        },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
            },
        'default': {
            'handlers': ['database',],
            'level': 'DEBUG',
            'propagate': True,
            },
        }
}


#From cerberos
MAX_FAILED_LOGINS = 5
MEMORY_FOR_FAILED_LOGINS = 3600  #try again an hour later

#from django-passwords
PASSWORD_MIN_LENGTH = 8
PASSWORD_COMPLEXITY = { "DIGITS": 1, "UPPER": 1 }

#uncomment this or put in your local settings if you want to save rml
SAVERML = os.path.join(PROJECT_DIR,'latest.rml')

TRACK_AJAX_CALLS = True

TESTING = 'test' in sys.argv

try:
    from localsettings import *
except ImportError:
    pass

#recommendations for security from: http://django-secure.readthedocs.org/en/v0.1.2/
SECURE_SSL_REDIRECT = False
SECURE_HSTS_SECONDS = 24*24*3600*30
SECURE_HSTS_SECONDS_INCLUDE_SUBDOMAINDS = True
SECURE_FRAME_DENY = True
SECURE_CONTENT_TYPE_NOSNIFF = True
SECURE_BROWSER_XSS_FILTER = True
SESSION_COOKIE_SECURE = False
SESSION_COOKIE_HTTPONLY = True
SESSION_COOKIE_AGE = 3600

#avoids hourly logout when you're working
SESSION_SAVE_EVERY_REQUEST = True

from django.contrib import messages
MESSAGE_TAGS = {
    messages.DEBUG: 'alert-debug',
    messages.INFO: 'alert-info',
    messages.SUCCESS: 'alert-success',
    messages.WARNING: 'alert-warning',
    messages.ERROR: 'alert-danger error'
    }



DASHBOARD_TITLE = CLIENT_PROJECT + ' Dashboard'
THUMBNAIL_SIZES = (
    (120,80),
    (240,160),
    (320,240),
    )


from datetime import date
DATE_EARLY = date(2013,1,1)  #before anything in your accounts system
DATE_LATE = date(2099,1,1)  #after anything in your accounts system



from accountifie.middleware.docengine import getCurrentRequest
import django.template
from django.shortcuts import render

def dropdown_button(current, status, id, others, action):
    t = django.template.loader.get_template('dropdownbutton.html')
    context = django.template.RequestContext(getCurrentRequest())
    context['current'] = current
    context['status'] = status
    context['id'] = id
    context['others'] = dict((x[1],x[0]) for x in others)
    context['action'] = action
    return django.template.loader.render_to_string('dropdownbutton.html',  context)




'''
Simply pass 'SSL':True as a keyword argument from a url config and the user will
get redirected the same URL but on the https protocol.

E.g.
    urlpatterns = patterns('some_site.some_app.views', (r'^accounts/login/$','test_secure',{'SSL':True}), )
    
When you want to stop the secure session (e.g. when user logs out) just set
'SSL':True.

E.g.
    urlpatterns = patterns('some_site.some_app.views', (r'^accounts/logout/$','test_secure',{'SSL':False}), )

Original version:
    http://www.djangosnippets.org/snippets/85/

'''


__license__ = "Python"
__copyright__ = "Copyright (C) 2007, Stephen Zabel"
__author__ = "Stephen Zabel - sjzabel@gmail.com"
__contributors__ = "Jay Parlar - parlar@gmail.com"

from django.conf import settings
from django.http import HttpResponseRedirect, HttpResponsePermanentRedirect

SSL = getattr(settings, 'DOCENGINE_SSL', 'SSL')
SECURE_PATHS = getattr(settings, 'DOCENGINE_SECURE_PATHS', ('/admin/', settings.LOGIN_URL))
HTTPS_SUPPORT = getattr(settings, 'DOCENGINE_HTTPS_SUPPORT', True)

class SSLRedirect:
    def process_view(self, request, view_func, view_args, view_kwargs):
        if SSL in view_kwargs:
            desiredSSL = view_kwargs[SSL]
            del view_kwargs[SSL]
        else:
            desiredSSL = request.user.is_authenticated()
        """
        if not settings.DEVELOP:
            existingSSL = self._is_secure(request)
            if existingSSL != desiredSSL and HTTPS_SUPPORT:
                return self._redirect(request, desiredSSL)
        """
        
    def _is_secure(self, request):
        if request.is_secure():
            return True

        #Handle the Webfaction case until this gets resolved in the request.is_secure()
        if 'HTTP_X_FORWARDED_SSL' in request.META:
            return request.META['HTTP_X_FORWARDED_SSL'] == 'on'

        return False

    def _redirect(self, request, secure):
        protocol = secure and "https" or "http"
        newurl = "%s://%s%s" % (protocol,request.get_host(),request.get_full_path())
        if settings.DEBUG and request.method == 'POST':
            raise RuntimeError("""Django can't perform a SSL redirect while maintaining POST data.
           Please structure your views so that redirects only occur during GETs.""")

        return HttpResponsePermanentRedirect(newurl)



class SecureRequiredMiddleware(object):

    def __init__(self):
        self.paths = SECURE_PATHS 
        self.enabled = self.paths and HTTPS_SUPPORT and not settings.DEVELOP

    def process_request(self, request):
        if self.enabled and not request.is_secure():
            for path in self.paths:
                request_url = request.build_absolute_uri(request.get_full_path())
                if request.get_full_path().startswith(path) or \
                        path.startswith('http') and request_url.startswith(path): 
                        # just in case the paths are absolute URLs
                    secure_url = request_url.replace('http://', 'https://')
                    return HttpResponsePermanentRedirect(secure_url)
        return None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0007_auto_20180510_2236'),
    ]

    operations = [
        migrations.AlterField(
            model_name='counterparty',
            name='id',
            field=models.CharField(max_length=30, serialize=False, primary_key=True, validators=[django.core.validators.RegexValidator(b'^[0-9a-zA-Z]*$', b'Only alphanumeric characters are allowed. In particular.. no spaces.')]),
        ),
    ]



#command to populate calendar
import time
from datetime import date
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):

    help = 'Ensure calendar tables are populated between two dates'
    args = ''

    def handle(self, *app_labels, **options):
        verbose = False
        started = time.clock()
        from accountifie import cal
            
        from_date = None
        to_date = None
            
        cal.populate(from_date=from_date, to_date=to_date, verbose=verbose)
        finished = time.clock()
        print("populated calendar in %0.1f seconds" % (finished - started))





# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '__first__'),
    ]

    operations = [
        migrations.CreateModel(
            name='Alias',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=100)),
                ('display_as', models.CharField(max_length=100)),
                ('context', models.ForeignKey(blank=True, to='gl.Company', null=True, on_delete=models.CASCADE)),
            ],
        ),
        migrations.CreateModel(
            name='Config',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=20)),
            ],
        ),
        migrations.CreateModel(
            name='Variable',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('key', models.CharField(unique=True, max_length=100)),
                ('value', models.CharField(unique=True, max_length=100)),
            ],
        ),
        migrations.AddField(
            model_name='config',
            name='reporting',
            field=models.ForeignKey(to='environment.Variable', on_delete=models.CASCADE),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-12-21 12:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0005_auto_20171220_2138'),
    ]

    operations = [
        migrations.AlterField(
            model_name='counterparty',
            name='id',
            field=models.CharField(max_length=30, primary_key=True, serialize=False),
        ),
    ]



# -*- coding: utf-8 -*-


from django.db import migrations, models
import colorful.fields


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.CharField(help_text=b'Unique short code or number', max_length=20, serialize=False, primary_key=True)),
                ('path', models.CharField(help_text=b'dotted-path reference to show location on balance sheet', unique=True, max_length=100, db_index=True)),
                ('ordering', models.IntegerField(default=0)),
                ('display_name', models.CharField(help_text=b'the name you would use on a report title', max_length=100, blank=True)),
                ('role', models.CharField(help_text=b'used in P&L queries and for picklists', max_length=10, choices=[(b'asset', b'asset'), (b'liability', b'liability'), (b'income', b'income'), (b'expense', b'expense'), (b'capital', b'capital')])),
            ],
            options={
                'ordering': ('id',),
                'db_table': 'gl_account',
            },
        ),
        migrations.CreateModel(
            name='Company',
            fields=[
                ('id', models.CharField(max_length=5, serialize=False, primary_key=True)),
                ('name', models.CharField(max_length=50)),
                ('cmpy_type', models.CharField(max_length=10, choices=[(b'ALO', b'Standalone'), (b'CON', b'Consolidation')])),
                ('color_code', colorful.fields.RGBColorField(blank=True)),
                ('subs', models.ManyToManyField(related_name='_company_subs_+', to='gl.Company', blank=True)),
            ],
            options={
                'db_table': 'gl_company',
                'verbose_name_plural': 'Companies',
            },
        ),
        migrations.CreateModel(
            name='Counterparty',
            fields=[
                ('id', models.CharField(max_length=12, serialize=False, primary_key=True)),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'db_table': 'gl_counterparty',
                'verbose_name_plural': 'Counterparties',
            },
        ),
        migrations.CreateModel(
            name='Department',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=25)),
            ],
            options={
                'db_table': 'gl_department',
            },
        ),
        migrations.CreateModel(
            name='DepreciationPolicy',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('depreciation_period', models.PositiveIntegerField()),
                ('cap_account', models.ForeignKey(related_name='deppolicy_acct', to='gl.Account', on_delete=models.CASCADE)),
                ('depreciation_account', models.ForeignKey(related_name='deppolicy_depacct', to='gl.Account', on_delete=models.CASCADE)),
            ],
            options={
                'db_table': 'gl_depreciationpolicy',
            },
        ),
        migrations.CreateModel(
            name='Employee',
            fields=[
                ('employee_id', models.IntegerField(serialize=False, primary_key=True)),
                ('ach', models.CharField(max_length=200, null=True)),
                ('usr_disable', models.BooleanField()),
                ('e_mail', models.CharField(max_length=200, null=True)),
                ('role', models.CharField(max_length=200, null=True)),
                ('employee_name', models.CharField(max_length=200, null=True)),
                ('p_card', models.CharField(max_length=200, null=True)),
                ('department', models.ForeignKey(to='gl.Department', on_delete=models.CASCADE)),
            ],
            options={
                'db_table': 'gl_employee',
            },
        ),
        migrations.CreateModel(
            name='ExternalAccount',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('label', models.CharField(max_length=10)),
                ('name', models.CharField(max_length=100)),
                ('company', models.ForeignKey(to='gl.Company', on_delete=models.CASCADE)),
                ('counterparty', models.ForeignKey(to='gl.Counterparty', on_delete=models.CASCADE)),
                ('gl_account', models.ForeignKey(to='gl.Account', on_delete=models.CASCADE)),
            ],
            options={
                'db_table': 'gl_externalaccount',
            },
        ),
        migrations.CreateModel(
            name='ExternalBalance',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date', models.DateField(db_index=True)),
                ('balance', models.DecimalField(max_digits=11, decimal_places=2)),
                ('comment', models.CharField(max_length=100, null=True, blank=True)),
                ('account', models.ForeignKey(to='gl.Account', on_delete=models.CASCADE)),
                ('company', models.ForeignKey(to='gl.Company', on_delete=models.CASCADE)),
            ],
            options={
                'db_table': 'gl_externalbalance',
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.CharField(max_length=12, serialize=False, primary_key=True)),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'db_table': 'gl_project',
                'verbose_name_plural': 'Projects',
            },
        ),
        migrations.CreateModel(
            name='TranLine',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('amount', models.DecimalField(max_digits=11, decimal_places=2)),
                ('tags', models.CharField(max_length=200, blank=True)),
                ('account', models.ForeignKey(to='gl.Account', on_delete=models.CASCADE)),
                ('counterparty', models.ForeignKey(blank=True, to='gl.Counterparty', null=True, on_delete=models.CASCADE)),
            ],
            options={
                'db_table': 'gl_tranline',
            },
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date', models.DateField(db_index=True)),
                ('date_end', models.DateField(db_index=True, null=True, blank=True)),
                ('comment', models.CharField(max_length=100)),
                ('long_desc', models.CharField(max_length=200, null=True, blank=True)),
                ('object_id', models.PositiveIntegerField()),
                ('company', models.ForeignKey(to='gl.Company', on_delete=models.CASCADE)),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType', on_delete=models.CASCADE)),
            ],
            options={
                'db_table': 'gl_transaction',
            },
        ),
        migrations.AddField(
            model_name='tranline',
            name='transaction',
            field=models.ForeignKey(to='gl.Transaction', on_delete=models.CASCADE),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-12-31 22:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0005_auto_20171216_1804'),
    ]

    operations = [
        migrations.AddField(
            model_name='log',
            name='corrId',
            field=models.CharField(blank=True, max_length=20, null=True),
        ),
    ]




"""
Adapted with permission from ReportLab's DocEngine framework
"""


from django.db import connection


def query(stmt, *args):
    """Execute statement (optionally with parameters) and return results.
    
    Returns list of lists


    """

    cur = connection.cursor()
    if args:
        cur.execute(stmt, args)
    else:
        cur.execute(stmt)
    data = cur.fetchall()
    cur.close()
    return data


def query_as_set(stmt, *args):
    "For one-column query, return a set of values"
    data = query(stmt, *args)
    assert len(data[0]) == 1, 'query_one_column requires a one-column query!'
    return set([row[0] for row in data])




from django.conf.urls import url

from . import background_status
from . import views

urlpatterns = [
    url(r'background_task/status/(?P<task_id>[\w\-]+)/$', background_status),
    url(r'tasks/list/$', views.tasks_list),
]
